using System;

namespace Framework
{
	/// <summary>
	/// Represent min-max box.
	/// It is automatically maintained by given point array.
	/// </summary>
	public sealed class BoundingBox
	{
		private bool isMinMaxVaild;

		/// <summary>
		/// Number of points with maximum dimension
		/// </summary>
		private int maxDimCnt;

		/// <summary>
		/// Current maximum dimension
		/// </summary>
		private int maxDim;

        private Point3D maxPoint;

        public Point3D MaxPoint
        {
            get { return maxPoint; }
            set { maxPoint = value; }
        }
        private Point3D minPoint;

        public Point3D MinPoint
        {
            get { return minPoint; }
            set { minPoint = value; }
        }
        private Point3D midPoint;

        public Point3D MidPoint
        {
            get { return midPoint; }
            set { midPoint = value; }
        }
		private double maxDiag;

		private Point3D[] points;

		/// <summary>
		/// Creates bounding box that is automatically maintained with respect to given array.
		/// </summary>
		/// <param name="points"></param>
		public BoundingBox(Point3D[] points)
		{
			this.points = points;
            this.ValidateMinMax();
            this.ValidateMidDiag();
		}        

        /// <summary>
        /// Returns point array that is managed by curent bounding box.
        /// </summary>
        public Point3D[] Points
		{
			get
			{
				return this.points;
			}
		}

		/// <summary>
		/// Recomputes the BoundingBox from all points
		/// </summary>
		private void ValidateMinMax()
		{
			for (int i = 0; i < 3; i++)
			{
				minPoint[i] = Double.MaxValue;
				maxPoint[i] = Double.MinValue;
			}

			for (int i = 0; i < points.Length; i++) 
			{
				Point3D point = points[i];

				for(int dim = 0; dim < 3; dim++) 
				{
					minPoint[dim] = Math.Min(minPoint[dim], point[dim]);
					maxPoint[dim] = Math.Max(maxPoint[dim], point[dim]);
				}
			}
			isMinMaxVaild = true;
			this.ValidateMidDiag();
		}

		/// <summary>
		/// Recompute mid-point and max diagonal length.
		/// </summary>
		private void ValidateMidDiag()
		{
            midPoint = (minPoint + maxPoint) * 0.5;
            Point3D diag = maxPoint - minPoint;
			this.maxDiag = diag.Abs();
		}

		/// <summary> 
		/// Return volume diagonal of the bounding box.
		/// </summary> 
		public double MaxDiag  
		{ 
			get 
			{ 
				if(!isMinMaxVaild)  
					this.ValidateMinMax();  

				return maxDiag; 
			} 
		}

		/// <summary>
		/// Returns description of BoundignBox.
		/// </summary>
		public override string ToString()
		{
			string str = "BoundingBox: ";

			if (this.MinPoint == null || this.MaxPoint == null || this.MidPoint == null)
				return str += "undefined (there are no points)";

			str += "\n";
			str += "\t" + "- min point : " + this.MinPoint + "\n";
			str += "\t" + "- max point : " + this.MaxPoint + "\n";
			str += "\t" + "- mid point : " + this.MidPoint + "\n";
			str += "\t" + "- max diag  : " + this.MaxDiag + "\n";
			return str;
		}

	} // class BoundingBox
} // namespace
