﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Framework
{
    public struct Point3D
    {
        public const double EPSILON = double.Epsilon * 3;

        public double X;
        public double Y;
        public double Z;

        public double this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return X;
                    case 1: return Y;
                    case 2: return Z;
                    default: throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
                }
            }
            set
            {
                switch (index)
                {
                    case 0: X = value; break;
                    case 1: Y = value; break;
                    case 2: Z = value; break;
                    default: throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
                }
            }
        }

        public Point3D(double x, double y, double z)
        {
            X = x;
            Y = y;
            Z = z;
        }

        public Point3D(params double[] position)
        {
            if (position.Length < 3)
            {
                throw new Exception("Less than three parameters passed to Point3D constructor");
            }
            else
            {
                X = position[0];
                Y = position[1];
                Z = position[2];
            }
        }

        public double Abs()
        {
            return Math.Sqrt(X * X + Y * Y + Z * Z);
        }

        #region Overriden methods of the class Object

        override public string ToString()
        {
            return "Point3D: [" + X + "; " + Y + "; " + Z + "]";
        }

        public static new bool Equals(object a, object b)
        {
            if (a.GetType() != typeof(Point3D) || b.GetType() != typeof(Point3D))
                return false;

            return Math.Abs(((Point3D)a).X - ((Point3D)b).X) <= Point3D.EPSILON
                && Math.Abs(((Point3D)a).Y - ((Point3D)b).Y) <= Point3D.EPSILON
                && Math.Abs(((Point3D)a).Z - ((Point3D)b).Z) <= Point3D.EPSILON;
        }

        public override bool Equals(object obj)
        {
            return Point3D.Equals(this, obj);
        }

        public override int GetHashCode()
        {
            return this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Z.GetHashCode();
        }

        #endregion

        #region Operators

        /// <summary>
        /// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>True if the values of operands are equal, false otherwise.</returns>
        public static bool operator ==(Point3D a, Point3D b)
        {
            return Point3D.Equals(a, b);
        }

        /// <summary>
        /// The inequality operator (!=) returns false if its operands are equal, true otherwise.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>False if the values of operands are equal, true otherwise.</returns>
        public static bool operator !=(Point3D a, Point3D b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D operator +(Point3D a, Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D operator -(Point3D a, Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

        /// <summary>
        /// Sign inverse of a point.
        /// </summary>
        /// <param name="p">Operand.</param>
        /// <returns>Resulting point.</returns>
        public static Point3D operator -(Point3D p)
        {
            return (new Point3D(-p.X, -p.Y, -p.Z));
        }

        /// <summary>
        /// Multiplication by constant number.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">multiplicator.</param>
        /// <returns>Resulting point (m*p.x, m*p.y, m*p.z)</returns>
        public static Point3D operator *(Point3D p, double m)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }

        /// <summary>
        /// Multiplication by a scalar.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">Multiplicator.</param>
        /// <returns>Resulting point (m*p.x, m*p.y, m*p.z).</returns>
        public static Point3D operator *(double m, Point3D p)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }

        /// <summary>
        /// Division by a scalar.
        /// </summary>
        /// <param name="p">Point to be divided.</param>
        /// <param name="m">Divider.</param>
        /// <returns>Resulting point (p.x/m, p.y/m, p.z/m).</returns>
        public static Point3D operator /(Point3D p, double m)
        {
            return (new Point3D(p.X / m, p.Y / m, p.Z / m));
        }

        #endregion

        #region Operations

        /// <summary>
        /// Returns euclidean distance between two points.
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Euclidean distance between the points.</returns>
        public double DistanceTo(Point3D p)
        {
            return (Math.Sqrt((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z)));
        }

        /// <summary>
        /// Returns squared euclidean distance between two points
        /// (is slightly faster than the DistanceTo function as no square root is computed),
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Square euclidean distance between the points.</returns>
        public double SquareDistanceTo(Point3D p)
        {
            return ((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z));
        }

        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the dot product operation.</param>
        /// <returns>Scalar value of this.arg.</returns>
        public double DotProduct(Point3D arg)
        {
            return (this.X * arg.X + this.Y * arg.Y + this.Z * arg.Z);
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public Point3D CrossProduct(Point3D arg)
        {
            return (new Point3D(this.Y * arg.Z - this.Z * arg.Y, this.Z * arg.X - this.X * arg.Z, this.X * arg.Y - this.Y * arg.X));
        }

        #endregion

        public double dotProduct(Point3D arg)
        {
            return (this.X * arg.X + this.Y * arg.Y + this.Z * arg.Z);
        }
    }
}
