﻿using System;
//using System.Drawing;
using System.ComponentModel;
using SlimDX;
using SlimDXRenderSystem.Core;
using SlimDXRenderSystem.Direct3DBase;
using SlimDX.Direct3D9;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using SceneMgr = SlimDXRenderSystem.Managers.SceneManager;
using EffectMgr = SlimDXRenderSystem.Managers.EffectManager;
using Effect = SlimDXRenderSystem.Direct3DBase.Effect;
using Light = SlimDXRenderSystem.Core.Light;
using System.Runtime.Serialization;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Reflection;
using System.Windows;
using System.Windows.Media;

namespace SlimDXRenderSystem.Direct3D9
{
    /// <summary>
    /// Direct3D9 Renderer
    /// </summary>
    public class Renderer9 : Renderer
    {
        #region Fields
        private Direct3DEx d3d;
        private DeviceEx device;
        private PresentParameters presentParams;
        private Light globalLight, cursorLight;
        private MultisampleQuality multisample;
        private SlimDX.Direct3D9.Font font;
        private Sprite fontSprite;
        private Texture pointTex;
        private int width, height;
  
        #region Properties
        /// <summary>
        /// Returns Direct3D9 device.
        /// </summary>
        [Browsable(false)]
        public Device Device
        {
            get { return device; }
        }

        /// <summary>
        /// Shade mode of mesh.
        /// </summary>
        public override ShadeMode Shademode
        {
            set
            {
                device.SetRenderState(RenderState.ShadeMode, value);
            }
        }

        /// <summary>
        /// Fill mode of mesh.
        /// </summary>
        public override FillMode Fillmode
        {
            set
            {
                device.SetRenderState(RenderState.FillMode, value);
            }
        }

        /// <summary>
        /// Back face culling.
        /// </summary>
        public override Cull Culling
        {
            set
            {
                device.SetRenderState(RenderState.CullMode, value);
            }
        }

        /// <summary>
        /// Antialiasing multi sample type.
        /// </summary>
        public override MultisampleQuality Multisample
        {
            get { return multisample; }
            set
            {
                if (d3d.CheckDeviceMultisampleType(0, DeviceType.Hardware, Format.X8R8G8B8, true, (MultisampleType)value))
                {
                    multisample = value;
                    RenderMgr.Instance.Reset();
                }
            }
        }

        /// <summary>
        /// Set default lighting the scene
        /// </summary>
        public bool Lighting
        {
            set
            {
                device.SetRenderState(RenderState.Lighting, value);
            }
        }

        /// <summary>
        /// Color of the default ambient light
        /// </summary>
        public Color Ambient
        {
            set
            {
                device.SetRenderState(RenderState.Ambient, (value.A << 24) | (value.R << 16) | (value.G << 8) | value.B);
            }
        }

        public PresentParameters PresentParams
        {
            get
            {
                return presentParams;
            }
        }
        #endregion
        #endregion

        #region Private Methods
        /// <summary>
        /// Sets present parameters for Direct3D9 device.
        /// </summary>
        private void SetPresentParameters()
        {
            var dpiXProperty = typeof(SystemParameters).GetProperty("DpiX", BindingFlags.NonPublic | BindingFlags.Static);
            var dpiYProperty = typeof(SystemParameters).GetProperty("Dpi", BindingFlags.NonPublic | BindingFlags.Static);

            var dpiX = (int)dpiXProperty.GetValue(null, null);
            var dpiY = (int)dpiYProperty.GetValue(null, null);

            presentParams = new PresentParameters()
            {
                SwapEffect = SwapEffect.Discard,
                EnableAutoDepthStencil = true,
                AutoDepthStencilFormat = Format.D16,
                BackBufferFormat = Format.X8R8G8B8,
                BackBufferHeight = (int)(RenderMgr.Instance.RenderWindow.ActualHeight*dpiY/96)-3, // ???
                BackBufferWidth = (int)(RenderMgr.Instance.RenderWindow.ActualWidth*dpiX/96)-3, // ???
                /*BackBufferHeight = (int)RenderMgr.Instance.RenderWindow.Height, // ???
                BackBufferWidth = (int)RenderMgr.Instance.RenderWindow.D3dimage.PixelWidth, // ???*/
                Multisample = (MultisampleType)this.multisample
            };
            width = presentParams.BackBufferWidth;
            height = presentParams.BackBufferHeight;      
        }

        /// <summary>
        /// Sets default renderstates.
        /// </summary>
        private void SetRenderStates()
        {
            device.SetRenderState(RenderState.FillMode, RenderMgr.Instance.Fillmode);
            device.SetRenderState(RenderState.CullMode, RenderMgr.Instance.Culling);
            device.SetRenderState(RenderState.ZEnable, true);
            device.SetRenderState(RenderState.ShadeMode, RenderMgr.Instance.Shademode);
        }

        /// <summary>
        /// Creates global light.
        /// </summary>
        private void SetGlobalLight()
        {
            globalLight = SceneMgr.Instance.CreateLight("Default");
            globalLight.Diffuse = Colors.White;
            globalLight.Type = LightType.Directional;
            globalLight.Direction = new Vector3(-1, 0, 0);
            globalLight.Specular = Colors.White;
            globalLight.Ambient = Colors.Black;

            cursorLight = SceneMgr.Instance.CreateLight("Cursor");
            cursorLight.Diffuse = Colors.Green;
            cursorLight.Ambient = Colors.White;
            cursorLight.Type = LightType.Point;
            cursorLight.Position = Vector3.TransformCoordinate(new Vector3(0, 0, 0), SceneMgr.Instance.WorldMatrix);
            cursorLight.Specular = Colors.White;
            cursorLight.Radius = 0.2f;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Constructor.
        /// </summary>
        public Renderer9()
        {
            multisample = RenderMgr.Instance.Multisample;
        }

        /// <summary>
        /// Initialize device and sets all needed states.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();
            SetPresentParameters();
            d3d = new Direct3DEx();
            device = new DeviceEx(d3d, 0, DeviceType.Hardware, RenderMgr.Instance.RenderWindow.Hwnd.Handle, CreateFlags.HardwareVertexProcessing, presentParams);
            SetRenderStates();
            SetGlobalLight();
            //pointTex = Texture.FromFile(device, "point.bmp");
            try
            {
                Assembly assembly = this.GetType().Assembly;
                string assemblyName = assembly.ToString().Split(',')[0];
                Uri uri = new Uri(String.Format("pack://application:,,,/{0};component/Resources/Textures/point.bmp", assemblyName));
                pointTex = Texture.FromStream(device, Application.GetResourceStream(uri).Stream);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.StackTrace);
                throw;
            }
        }

        

        /// <summary>
        /// Render scene.
        /// </summary>
        public override void Render()
        {
            try
            {
                Effect effect = EffectMgr.Instance.SuperEffect;

                effect.SetParameter(Effect.VIEW, SceneMgr.Instance.Camera.ViewMatrix);
                effect.SetParameter(Effect.PROJECTION, SceneMgr.Instance.Camera.ProjectionMatrix);
                effect.SetParameter(Effect.WORLD, SceneMgr.Instance.WorldMatrix);
                effect.SetParameter(Effect.WORLDINVTRANSPOSE, SlimDX.Matrix.Transpose(SlimDX.Matrix.Invert(SceneMgr.Instance.WorldMatrix)));
                effect.SetParameter(Effect.WORLDVIEWPROJECTION, SceneMgr.Instance.WorldMatrix * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix);
                effect.SetParameter(Effect.CAMERAPOSITION, SceneMgr.Instance.Camera.Position);
                effect.SetParameter(Effect.LIGHT, globalLight.LightStruct);
                effect.SetParameter(Effect.LIGHT2, cursorLight.LightStruct);


                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, BackgroundColor, 1.0f, 0);
                device.BeginScene();

                int passes = ((Effect9)effect).Begin();

                for (int i = 0; i < passes; i++)
                {
                    ((Effect9)effect).BeginPass(i);
                    //Draw meshes
                    SceneMgr.Instance.DrawObjects();

                    ((Effect9)effect).EndPass();
                }


                ((Effect9)effect).End();
                
                //Draw text
                RenderText();
                
                device.EndScene();
                device.Present();
            }
            catch (SlimDXException)
            {
                RenderMgr.Instance.Reset();
            }
        }

        public override void RenderText()
        {
            if (fontSprite == null)
                fontSprite = new Sprite(device);
            if (font == null)
            {
                System.Drawing.Font fnt = new System.Drawing.Font(System.Drawing.FontFamily.GenericSansSerif, 12.0f);
                font = new SlimDX.Direct3D9.Font(device, fnt);
            }
            fontSprite.Begin(SpriteFlags.AlphaBlend | SpriteFlags.SortTexture);

            if (RenderMgr.Instance.RenderTextAction != null)
                RenderMgr.Instance.RenderTextAction.Invoke();

            fontSprite.End();
        }

        public override void DrawPoint(Vector3 pos)
        {
            fontSprite.Draw(pointTex, new Vector3(2, 2, 0), new Vector3(pos.X*width/2+width/2, -pos.Y*height/2+height/2, 0), new Color4(1,1,1));
        }

        public override void DrawText(float x, float y, string text)
        {
            font.DrawString(fontSprite, text, (int)(x*width/2+width/2), (int)(-y*height/2+height/2), new Color4(1, 1, 1));
        }

        public override void DrawText(int x, int y, string text)
        {
            font.DrawString(fontSprite, text, x, y, new Color4(1, 1, 1));
        }

        /// <summary>
        /// Reset device and set all needed states.
        /// </summary>
        public override void Reset()
        {
            SetPresentParameters();
            if (font != null)
            {
                font.Dispose();
                font = null;
            }
            if (fontSprite != null)
            {
                fontSprite.Dispose();
                fontSprite = null;
            }
            device.Reset(presentParams);
            SetRenderStates();
            //SetGlobalLight();
        }

        /// <summary>
        /// Capture screenshot of window frame
        /// </summary>
        /// <param name="name"></param>
        /// <param name="number"></param>
        public override void CaptureScreenshot(String name, int number)
        {
            CaptureScreenshot(name + "_" + number + ".bmp");            
        }

        /// <summary>
        /// Capture screenshot of window frame
        /// </summary>
        /// <param name="name"></param>
        /// <param name="number"></param>
        public override void CaptureScreenshot(String name)
        {
            Surface backbuffer = device.GetBackBuffer(0, 0);
            Surface.ToFile(backbuffer, name, ImageFileFormat.Bmp);
            backbuffer.Dispose();
        }

        /// <summary>
        /// Saves the position of the light to the file.
        /// </summary>
        /// <param name="file">File name of the light transformations used for saving.</param>
        public override void SaveLightTransformations(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                                             FileMode.Create,
                                             FileAccess.Write, FileShare.None);
            formatter.Serialize(stream, globalLight.Direction.X);
            formatter.Serialize(stream, globalLight.Direction.Y);
            formatter.Serialize(stream, globalLight.Direction.Z);
            formatter.Serialize(stream, globalLight.Position.X);
            formatter.Serialize(stream, globalLight.Position.Y);
            formatter.Serialize(stream, globalLight.Position.Z);
            formatter.Serialize(stream, globalLight.Diffuse.ScR);
            formatter.Serialize(stream, globalLight.Diffuse.ScG);
            formatter.Serialize(stream, globalLight.Diffuse.ScB);
            formatter.Serialize(stream, globalLight.Diffuse.ScA);
            formatter.Serialize(stream, globalLight.Ambient.ScR);
            formatter.Serialize(stream, globalLight.Ambient.ScG);
            formatter.Serialize(stream, globalLight.Ambient.ScB);
            formatter.Serialize(stream, globalLight.Ambient.ScA);
            formatter.Serialize(stream, globalLight.Specular.ScR);
            formatter.Serialize(stream, globalLight.Specular.ScG);
            formatter.Serialize(stream, globalLight.Specular.ScB);
            formatter.Serialize(stream, globalLight.Specular.ScA);
            stream.Close();
        }

        /// <summary>
        /// Loads the position of the light to the file.
        /// </summary>
        /// <param name="file">File name of the light transformations used for loading.</param>
        public override void LoadLightTransformations(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                          FileMode.Open,
                          FileAccess.Read,
                          FileShare.Read);
            float x = (float)formatter.Deserialize(stream);
            float y = (float)formatter.Deserialize(stream);
            float z = (float)formatter.Deserialize(stream);
            globalLight.Direction = new Vector3(x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            globalLight.Position = new Vector3(x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            float w = (float)formatter.Deserialize(stream);
            globalLight.Diffuse = Color.FromScRgb(w, x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            w = (float)formatter.Deserialize(stream);
            globalLight.Ambient = Color.FromScRgb(w, x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            w = (float)formatter.Deserialize(stream);
            globalLight.Specular = Color.FromScRgb(w, x, y, z);

            stream.Close();
        }

        /// <summary>
        /// Clear all resources.
        /// </summary>
        public override void Dispose()
        {
            if (d3d != null && !d3d.Disposed)
            {
                d3d.Dispose();
                d3d = null;
            }
            if (device != null && !device.Disposed)
            {
                device.Dispose();
                device = null;
            }
            if (font != null && !font.Disposed)
            {
                font.Dispose();
                font = null;
            }
            if (fontSprite != null && !fontSprite.Disposed)
            {
                fontSprite.Dispose();
                fontSprite = null;
            }
            base.Dispose();
        }
        #endregion
    }
}
