﻿using System;
using System.Windows.Media;
using System.Windows;
using SlimDX;
using SlimDXRenderSystem.Direct3DBase;

namespace SlimDXRenderSystem.Managers
{
    /// <summary>
    /// Render manager takes control of rendering scene using specificed Direct3D version.
    /// </summary>
    public class RenderManager : Core.Singleton<RenderManager>
    {
        #region Fields
        /// <summary>
        /// World X axis vector.
        /// </summary>
        public static readonly Vector3 WORLD_X_AXIS = Vector3.UnitX;

        /// <summary>
        /// World Y axis vector.
        /// </summary>
        public static readonly Vector3 WORLD_Y_AXIS = Vector3.UnitY;

        /// <summary>
        /// World Z axis vector.
        /// </summary>
        public static readonly Vector3 WORLD_Z_AXIS = Vector3.UnitZ;

        private static readonly D3DVersion DEF_D3DVERSION = D3DVersion.Direct3D9;
        private static readonly Color DEF_AMBIENT = Colors.Gray;
        private static readonly ShadeMode DEF_SHADEMODE = ShadeMode.Gouraud;
        private static readonly FillMode DEF_FILLMODE = FillMode.Solid;
        private static readonly Cull DEF_CULL = Cull.None;
        //private static readonly MultisampleQuality DEF_MULTISAMPLE = MultisampleQuality.Samples_8;
        private static readonly MultisampleQuality DEF_MULTISAMPLE = MultisampleQuality.None;
        private const float DEF_AMBIENT_INTENSITY = 0.25f;
        private const bool DEF_LIGHTING = true;
        private const string DEF_EFFECT_RESOURCE = "MeshBrowser.Resources.Effects.default.fx";
        //private const string DEF_EFFECT_RESOURCE = "default.fx";

        private Window mainWindow;
        private SlimDXControl renderWindow;
        private D3DVersion d3dVersion;

        private Renderer renderer;


        //Global setting variables
        private bool lighting;
        private System.Windows.Media.Color ambientLight;
        private float ambientIntensity;
        private ShadeMode shademode;
        private FillMode fillmode;
        private Cull culling;
        private MultisampleQuality multisample;
        private Action renderTextAction = null;
        private bool wizardShown = true;

        

        #region Properties
        /// <summary>
        /// Main form for event handling.
        /// </summary>
        public Window MainWindow
        {
            get { return mainWindow; }
            set { mainWindow = value; }
        }

        public bool WizardShown
        {
            get { return wizardShown; }
            set { wizardShown = value; }
        }

        /// <summary>
        /// Main rendering control.
        /// </summary>
        public SlimDXControl RenderWindow
        {
            get { return renderWindow; }
            set { renderWindow = value; }
        }

        /// <summary>
        /// Direct3D version used for rendering.
        /// </summary>
        public D3DVersion D3dVersion
        {
            get { return d3dVersion; }
            set { d3dVersion = value; }
        }

        /// <summary>
        /// Direct3D renderer.
        /// </summary>
        public Renderer Renderer
        {
            get { return renderer; }
        }
        

        public Action RenderTextAction
        {
            get { return renderTextAction; }
            set { renderTextAction = value; }
        }

        //Global setting properties
        /// <summary>
        /// Toogle Lighting
        /// </summary>
        public bool Lighting
        {
            get { return lighting; }
            set
            {
                lighting = value;
                EffectManager.Instance.SetParameter(Effect.LIGHTING, lighting);
            }
        }

        /// <summary>
        /// Ambient light color.
        /// </summary>
        public System.Windows.Media.Color AmbientLight
        {
            get { return ambientLight; }
            set
            {
                ambientLight = value;
                EffectManager.Instance.SetParameter(Effect.AMBIENTCOLOR, ambientLight);
            }
        }

        /// <summary>
        /// Ambient light intensity.
        /// </summary>
        public float AmbientIntensity
        {
            get { return ambientIntensity; }
            set
            {
                if (value > 2.0f)
                    ambientIntensity = 2.0f;
                else if (value < 0.0f)
                    ambientIntensity = 0.0f;
                else
                    ambientIntensity = value;
                EffectManager.Instance.SetParameter(Effect.AMBIENTINTENSITY, ambientIntensity);
            }
        }

        /// <summary>
        /// Mesh shading.
        /// </summary>
        public ShadeMode Shademode
        {
            get { return shademode; }
            set
            {
                shademode = value;
                renderer.Shademode = shademode;
            }
        }

        /// <summary>
        /// Fillmode.
        /// </summary>
        public FillMode Fillmode
        {
            get { return fillmode; }
            set
            {
                fillmode = value;
                renderer.Fillmode = fillmode;
            }
        }

        /// <summary>
        /// Backface culling.
        /// </summary>
        public Cull Culling
        {
            get { return culling; }
            set
            {
                culling = value;
                renderer.Culling = culling;
            }
        }

        /// <summary>
        /// Antialiasing multisample quality.
        /// </summary>
        public MultisampleQuality Multisample
        {
            get { return multisample; }
            set
            {
                renderer.Multisample = value;
                multisample = renderer.Multisample;
            }
        }
        #endregion
        #endregion
        #region Public Methods
        /// <summary>
        /// Initialize RenderManager and prepare it for rendering.
        /// </summary>
        /// <param name="mainWindow">Main form for event handling.</param>
        /// <param name="renderWindow">Rendering control used for rendering.</param>
        public void Initialize(Window mainWindow, SlimDXControl renderWindow)
        {
            /*Default values*/
            d3dVersion = DEF_D3DVERSION;
            lighting = DEF_LIGHTING;
            ambientLight = DEF_AMBIENT;
            ambientIntensity = DEF_AMBIENT_INTENSITY;
            shademode = DEF_SHADEMODE;
            fillmode = DEF_FILLMODE;
            culling = DEF_CULL;
            multisample = DEF_MULTISAMPLE;

            this.mainWindow = mainWindow;
            this.renderWindow = renderWindow;

            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    renderer = new Direct3D9.Renderer9();
                    break;
            }
            renderer.Initialize();
            EffectManager.Instance.LoadEffectFromResource(DEF_EFFECT_RESOURCE);
        }

        /// <summary>
        /// Render scene.
        /// </summary>
        public void Render()
        {
            renderer.Render();
            TimeManager.Instance.Update();
        }

        public void DrawPoint(Vector3 pos)
        {
            renderer.DrawPoint(pos);
        }

        public void DrawText(float x, float y, string text)
        {
            renderer.DrawText(x, y, text);
        }

        public void DrawText(int x, int y, string text)
        {
            renderer.DrawText(x, y, text);
        }

        /// <summary>
        /// Reset all stuff.
        /// </summary>
        public void Reset()
        {
            EffectManager.Instance.Dispose();
            renderer.Reset();
            EffectManager.Instance.LoadEffectFromResource(DEF_EFFECT_RESOURCE);
        }

        public void CaptureScreenshot(string filename, int framenum)
        {
            renderer.CaptureScreenshot(filename, framenum);
        }

        public void CaptureScreenshot(string filename)
        {
            renderer.CaptureScreenshot(filename);
        }


        /// <summary>
        /// Clear all resources.
        /// </summary>
        public void Dispose()
        {
            if (renderer != null && !renderer.Disposed)
            {
                renderer.Dispose();
                renderer = null;
            }
        }
        #endregion
    }
}
