﻿using System;
using System.Collections.Generic;
using SlimDX;
using SlimDXRenderSystem.Direct3DBase;
using SlimDXRenderSystem.Core;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

namespace SlimDXRenderSystem.Managers
{
    public class SceneManager : Core.Singleton<SceneManager>
    {
        #region Fields
        private Dictionary<String, Group> sceneGroups;
        private Dictionary<String, Object3D> sceneObjects;
        private Dictionary<String, Camera> sceneCameras;
        private Dictionary<String, Light> sceneLights;
        private Light activeLight;
        private Light cursorLight;

        
        private Camera activeCamera;
        private Group undefGrp = new Group("Undefined");

        private Object3D cursor;

        public Object3D Cursor
        {
            get { return cursor; }
            set { cursor = value; }
        }

        private Matrix worldMatrix = Matrix.Identity;

        #region Properties
        /// <summary>
        /// Returns active camera.
        /// </summary>
        public Camera Camera
        {
            get { return activeCamera; }
        }

        public Light Light
        {
            get { return activeLight; }
        }

        public Light CursorLight
        {
            get { return cursorLight; }
            set { cursorLight = value; }
        }

        public Dictionary<String, Group> SceneGroups
        {
            get { return sceneGroups; }
        }
        public Dictionary<String, Camera> SceneCameras
        {
            get { return sceneCameras; }
        }
        public Dictionary<String, Light> SceneLights
        {
            get { return sceneLights; }
        }

        /// <summary>
        /// Global world matrix.
        /// </summary>
        public Matrix WorldMatrix
        {
            get { return worldMatrix; }
            set
            {
                worldMatrix = value;
                if (EffectManager.Instance.SuperEffect != null)
                    EffectManager.Instance.SetParameter(Effect.WORLD, worldMatrix);
            }
        }

        #endregion
        #endregion

        #region Public Methods

        public void SetWorldTransform(Matrix mat)
        {
            EffectManager.Instance.SetParameter(Effect.WORLD, mat);
            EffectManager.Instance.SetParameter(Effect.WORLDVIEWPROJECTION, mat * Camera.ViewMatrix * Camera.ProjectionMatrix);
        }

        public void Initialize()
        {
            sceneGroups = new Dictionary<string, Group>();
            sceneObjects = new Dictionary<string, Object3D>();

            sceneGroups.Add("Undefined", undefGrp);
        }

        /// <summary>
        /// Creates new object and return its pointer.
        /// </summary>
        /// <param name="name">Name of the object.</param>
        /// <returns>Pointer to new object.</returns>
        public Object3D CreateObject(String name)
        {
            if (sceneObjects.ContainsKey(name))
                return null; 

            Object3D newObject = new Object3D(name);

            if (name != "Cursor")
                sceneObjects.Add(name, newObject);
            else
                cursor = newObject;

            undefGrp.AddObject(newObject);
            return newObject;
        }

        public Group CreateGroup(String name, Movable m)
        {
            if (sceneGroups == null)
                sceneGroups = new Dictionary<string,Group>();
            if (sceneGroups.ContainsKey(name))
                return null;

            Group newGroup = new Group(name);

            sceneGroups.Add(name, newGroup);

            if (m != null)
            {
                newGroup.Position = m.Position;
                newGroup.RotationMatrix = m.RotationMatrix;
                newGroup.CalcMatrices();
            }

            return newGroup;
        }

        /// <summary>
        /// Adds given object into scene .
        /// </summary>
        /// <param name="name">Object.</param>
        public void AddObject(Object3D obj)
        {
            if (sceneObjects == null)
                sceneObjects = new Dictionary<string, Object3D>();
            if (sceneObjects.ContainsKey(obj.Name))
                sceneObjects[obj.Name] = obj; //Sets the old object to the new one

            if (obj.Group == null)
                undefGrp.AddObject(obj);

            sceneObjects.Add(obj.Name, obj);
        }

        /// <summary>
        /// Detach object from scene node and return.
        /// </summary>
        /// <param name="name">Name of the object to detach.</param>
        /// <returns>Detached object.</returns>
        public Object3D DetachObject(String name)
        {
            Object3D temp = sceneObjects[name];
            sceneObjects.Remove(name);
            return temp;
        }

        /// <summary>
        /// Return main object in scene with specified name.
        /// </summary>
        /// <param name="name">Name of object to get.</param>
        /// <returns>Pointer to found object, null if there is no object with this name.</returns>
        public Object3D GetObject(String name)
        {
            if (sceneObjects.ContainsKey(name))
                return sceneObjects[name];
            else
                return null;
        }

        /// <summary>
        /// This method will find object in entire scene.
        /// </summary>
        /// <param name="name">Name of object to find.</param>
        /// <returns>Pointer to found object, null if there is no object with this name.</returns>
        public Object3D FindObject(String name)
        {
            if (sceneObjects != null)
            {
                if (sceneObjects.ContainsKey(name))
                    return sceneObjects[name];
                Object3D temp = null;
                foreach (Object3D obj in sceneObjects.Values)
                {
                    temp = obj.FindChild(name);
                    if (temp != null)
                        return temp;
                }
                return temp;
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Creates new camera and set as active.
        /// </summary>
        /// <param name="name">Name of the camera object.</param>
        /// <returns>Pointer to created camera.</returns>
        public Camera CreateCamera(String name)
        {
            if (sceneCameras == null)
                sceneCameras = new Dictionary<string, Camera>();
            if (sceneCameras.ContainsKey(name))
                return null; 

            Camera cam = new Camera(name);
            cam.LookAt(new Vector3(0.0f, 0.0f, 1.0f), new Vector3(0.0f, 0.0f, 0.0f), Vector3.UnitY);
            sceneCameras.Add(name, cam);
            activeCamera = cam;
            return cam;
        }

        /// <summary>
        /// Find camera with specified name and set as active.
        /// </summary>
        /// <param name="name">Name of camera object.</param>
        public void SelectCamera(String name)
        {
            Camera temp = sceneCameras[name];
            if (temp == null)
                return;
            activeCamera = temp;
        }

        /// <summary>
        /// Remove camera with specified name from scene.
        /// </summary>
        /// <param name="name">Name of camera object.</param>
        /// <returns>If the camera was found and removed.</returns>
        public bool RemoveCamera(String name)
        {
            return sceneCameras.Remove(name);
        }

        public Light CreateLight(String name)
        {
            if (sceneLights == null)
                sceneLights = new Dictionary<string, Light>();
            if (sceneLights.ContainsKey(name))
                return sceneLights[name];
            Light light = new Light(name);
            sceneLights.Add(name, light);

            if (name != "Cursor")
            {
                if (activeLight == null)
                    activeLight = light;
            }
            else
                cursorLight = light;

            return light;
        }

        /// <summary>
        /// Draw the scene and all objects witch are visible.
        /// </summary>
        public void DrawObjects()
        {
            if (sceneObjects != null)
            {
                foreach (Object3D obj in sceneObjects.Values)
                {
                    obj.Draw();
                }
                
            }

            if (cursor != null)
            {                
                cursor.Draw();
            }
        }

        public void SaveWorld()
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream("world.bin",
                                             FileMode.Create,
                                             FileAccess.Write, FileShare.None);

            formatter.Serialize(stream, worldMatrix);

            foreach (Group g in sceneGroups.Values)
            {
                formatter.Serialize(stream, g.Position);
                formatter.Serialize(stream, g.RotationMatrix);
                formatter.Serialize(stream, g.Scale);
            }

            stream.Close();

            Camera.SavePosition("camera.bin");
            RenderManager.Instance.Renderer.SaveLightTransformations("light.bin");
        }

        public void LoadWorld()
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream("world.bin",
                          FileMode.Open,
                          FileAccess.Read,
                          FileShare.Read);

            worldMatrix = (Matrix)formatter.Deserialize(stream);

            foreach (Group g in sceneGroups.Values)
            {
                g.Position = (Vector3)formatter.Deserialize(stream);
                g.RotationMatrix = (Matrix)formatter.Deserialize(stream);
                g.Scale = (Vector3)formatter.Deserialize(stream);
            }

            stream.Close();

            Camera.LoadPosition("camera.bin");
            RenderManager.Instance.Renderer.LoadLightTransformations("light.bin");
        }

        /// <summary>
        /// Removes all objects and cameras.
        /// </summary>
        public void ClearScene()
        {
            if (sceneObjects != null)
            {
                foreach (Object3D obj in sceneObjects.Values)
                {
                    obj.Dispose();
                }
                sceneObjects.Clear();
            }
            if (sceneGroups != null)
            {
                foreach (Group grp in sceneGroups.Values)
                {
                    grp.Dispose();
                }
                sceneGroups.Clear();
            }
            /*if (sceneCameras != null)
                sceneCameras.Clear();
            activeCamera = null;*/
        }

        /// <summary>
        /// Dispose scene manager.
        /// </summary>
        public void Dispose()
        {
            ClearScene();
        }
        #endregion
    }
}
