struct Light
{
	float3 dir;                     // world space direction
	float3 pos;                     // world space position
	float4 ambient;
	float4 diffuse;
	float4 specular;
	float spotInnerCone;            // spot light inner cone (theta) angle
	float spotOuterCone;            // spot light outer cone (phi) angle
	float radius;                   // applies to point and spot lights only
};

struct Material
{
	float4 ambient;
	float4 diffuse;
	float4 emissive;
	float4 specular;
	float shininess;
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

float4x4 worldMatrix : WORLD;
float4x4 worldInverseTransposeMatrix : WORLDINVTRANSPOSE;
float4x4 viewMatrix : VIEW;
float4x4 projMatrix : PROJECTION;
float4x4 worldViewProjectionMatrix : WORLDVIEWPROJECTION;

float3 cameraPos : CAMERAPOSITION;

float4 globalAmbient : AMBIENTCOLOR;
float ambientIntensity : AMBIENTINTENSITY;

bool lighting : LIGHTING;

Light light : LIGHT;
Light light2 : LIGHT2;
Material material : MATERIAL;

//-----------------------------------------------------------------------------
// Vertex Shaders.
//-----------------------------------------------------------------------------

struct VS_INPUT
{
	float3 position : POSITION;
	float3 normal : NORMAL;
	float4 color : COLOR0;
};

struct VS_OUTPUT_DIR
{
	float4 position : POSITION;
	float3 halfVector : TEXCOORD0;
	float3 lightDir : TEXCOORD1;
	float3 normal : TEXCOORD2;
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

struct VS_OUTPUT_POINT
{
	float4 position : POSITION;
	float3 viewDir : TEXCOORD0;
	float3 lightDir : TEXCOORD1;
	float3 normal : TEXCOORD2;
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

struct VS_OUTPUT_SPOT
{
	float4 position : POSITION;
	float3 viewDir : TEXCOORD0;
	float3 lightDir : TEXCOORD1;
	float3 spotDir : TEXCOORD2;
	float3 normal : TEXCOORD4;
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

struct VS_OUTPUT_DIR_POINT
{
	float4 position : POSITION;
	float3 viewDir : TEXCOORD0;
	float3 lightDir : TEXCOORD1;
	float3 lightDir2 : TEXCOORD2;
	float3 normal : TEXCOORD3;
	float3 halfVector : TEXCOORD4;
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

VS_OUTPUT_DIR_POINT VS_DirPointLighting(VS_INPUT IN)
{

	VS_OUTPUT_DIR_POINT OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), worldMatrix).xyz;
		float3 viewDir = cameraPos - worldPos;

		OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.lightDir = -light.dir;
	OUT.lightDir2 = (light2.pos - worldPos) / light2.radius;
	OUT.halfVector = normalize(normalize(OUT.lightDir) + normalize(viewDir));
	OUT.normal = mul(IN.normal, (float3x3)worldInverseTransposeMatrix);
	OUT.diffuse = IN.color * material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
	OUT.viewDir = cameraPos - worldPos;

	return OUT;
}

VS_OUTPUT_DIR VS_DirLighting(VS_INPUT IN)
{

	VS_OUTPUT_DIR OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), worldMatrix).xyz;
	float3 viewDir = cameraPos - worldPos;
		
	OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.lightDir = -light.dir;
	OUT.halfVector = normalize(normalize(OUT.lightDir) + normalize(viewDir));
	OUT.normal = mul(IN.normal, (float3x3)worldInverseTransposeMatrix);
	OUT.diffuse = IN.color * material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
        
	return OUT;
}

VS_OUTPUT_POINT VS_PointLighting(VS_INPUT IN)
{
	VS_OUTPUT_POINT OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), worldMatrix).xyz;
    
	OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.viewDir = cameraPos - worldPos;
	OUT.lightDir = (light.pos - worldPos) / light.radius;
	OUT.normal = mul(IN.normal, (float3x3)worldInverseTransposeMatrix);
	OUT.diffuse = IN.color * material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
	
	return OUT;
}

VS_OUTPUT_SPOT VS_SpotLighting(VS_INPUT IN)
{
    VS_OUTPUT_SPOT OUT;
    
    float3 worldPos = mul(float4(IN.position, 1.0f), worldMatrix).xyz;
	       
    OUT.position = mul(float4(IN.position, 1.0f), worldViewProjectionMatrix);
	OUT.viewDir = cameraPos - worldPos;
	OUT.lightDir = (light.pos - worldPos) / light.radius;
    OUT.spotDir = light.dir;
    OUT.normal = mul(IN.normal, (float3x3)worldInverseTransposeMatrix);
    OUT.diffuse = IN.color * material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
       
    return OUT;
}

//-----------------------------------------------------------------------------
// Pixel Shaders.
//-----------------------------------------------------------------------------

float4 PS_DirPointLighting(VS_OUTPUT_DIR_POINT IN) : COLOR
{
	float3 n = normalize(IN.normal);
	float3 h = normalize(IN.halfVector);
	float3 l = normalize(IN.lightDir);

	float nDotL = saturate(dot(n, l));
	float nDotH = saturate(dot(n, h));
	float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
	float4 color;

	if (lighting)
	{
		color = (IN.diffuse * material.ambient * (globalAmbient + light.ambient)) +
			(IN.diffuse * nDotL) + (IN.specular * power);
	}
	else
	{
		color = IN.diffuse;
	}

	//float atten = 1.0f;
	float atten = saturate(1.0f - dot(IN.lightDir2, IN.lightDir2));

		float3 l2 = normalize(IN.lightDir2);
		float3 v = normalize(IN.viewDir);
		h = normalize(l2 + v);

		nDotL = saturate(dot(n, l2));
	nDotH = saturate(dot(n, h));
	power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);

	color += (material.ambient * (globalAmbient + (atten * light2.ambient))) +
		(light2.diffuse * nDotL * atten) + (light2.specular * power * atten);

	//color = (0.5f, 0.6f, 0.7f, 0.5f);

	return color;

}

float4 PS_DirLighting(VS_OUTPUT_DIR IN) : COLOR
{
    float3 n = normalize(IN.normal);
    float3 h = normalize(IN.halfVector);
    float3 l = normalize(IN.lightDir);
    
    float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);   
	float4 color;

	if (lighting)
	{
		color = (IN.diffuse * material.ambient * (globalAmbient + light.ambient)) +
                   (IN.diffuse * nDotL) + (IN.specular * power);
	}
	else
	{
		color = IN.diffuse;
	}

	return color;
}

float4 PS_PointLighting(VS_OUTPUT_POINT IN) : COLOR
{
    float atten = saturate(1.0f - dot(IN.lightDir, IN.lightDir));

	float3 n = normalize(IN.normal);
    float3 l = normalize(IN.lightDir);
    float3 v = normalize(IN.viewDir);
    float3 h = normalize(l + v);
    
    float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
    
	float4 color = (material.ambient * (globalAmbient + (atten * light.ambient))) +
                   (IN.diffuse * nDotL * atten) + (IN.specular * power * atten);
                   
	return color;
}

float4 PS_SpotLighting(VS_OUTPUT_SPOT IN) : COLOR
{
    float atten = saturate(1.0f - dot(IN.lightDir, IN.lightDir));
    
	float3 l = normalize(IN.lightDir);
    float2 cosAngles = cos(float2(light.spotOuterCone, light.spotInnerCone) * 0.5f);
    float spotDot = dot(-l, normalize(IN.spotDir));
    float spotEffect = smoothstep(cosAngles[0], cosAngles[1], spotDot);
    
    atten *= spotEffect;
                                
    float3 n = normalize(IN.normal);
	float3 v = normalize(IN.viewDir);
	float3 h = normalize(l + v);
    
    float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
    
    float4 color = (material.ambient * (globalAmbient + (atten * light.ambient))) +
                   (IN.diffuse * nDotL * atten) + (IN.specular * power * atten);
    
	return color;
}

//-----------------------------------------------------------------------------
// Techniques.
//-----------------------------------------------------------------------------

/*technique PerPixelDirectionalLighting2
{
	pass
	{
		VertexShader = compile vs_2_0 VS_DirLighting();
		PixelShader = compile ps_2_0 PS_DirLighting();
	}
}*/

technique PerPixelDirectionalLighting
{
	pass
	{
		VertexShader = compile vs_2_0 VS_DirPointLighting();
		PixelShader = compile ps_2_0 PS_DirPointLighting();
	}
}
/*
technique PerPixelPointLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_PointLighting();
        PixelShader = compile ps_2_0 PS_PointLighting();
    }
}

technique PerPixelSpotLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_SpotLighting();
        PixelShader = compile ps_2_0 PS_SpotLighting();
    }
}*/