﻿using System;
using System.Drawing;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using SlimDX;
using D3D9 = SlimDX.Direct3D9;
using D3D10 = SlimDX.Direct3D10;
using Device9 = SlimDX.Direct3D9.Device;
using Device10 = SlimDX.Direct3D10.Device;
using Matrix = SlimDX.Matrix;

namespace Zcu.Mve.ZMalik.SlimDXRenderer
{
    /// <summary>
    /// This enumeration defines Direct3D versions, that can be used for rendering.
    /// </summary>
    public enum D3DVersion
    {
        /// <summary>
        /// Defines Direct3D version to 9.
        /// </summary>
        Direct3D9,
        /// <summary>
        /// Defines Direct3D version to 10.
        /// </summary>
        Direct3D10
    }

    /// <summary>
    /// This class manage all rendering and graphic stuff like camera, lights and input.
    /// </summary>
    public sealed class GlobalManager
    {
        #region Fields
        private Setting9 setting9;
        private Setting10 setting10;

        private IntPtr hWnd;
        private Size clientSize;
        private D3DVersion d3dVersion = D3DVersion.Direct3D9;
        private Camera camera;

        private Color4 backColor;
        private int numScreenShots = 0;
        private float progress = 0.0f;
        private float lightAzimuth = 5.0f;
        private float lightElevation = 5.0f;
        private bool sameCon;
        private bool showColor;
        private bool showVector;

        private string colorAttrName;
        private string vectorAttrName;

        #region Direct3D9 Fields
        private D3D9.Direct3D d3d9;
        private Device9 device9;
        private Object3D9[] objects3D9;
        private D3D9.PresentParameters presentParameters;

        private D3D9.Light globalLight;

        #region Graphic Setting
        #endregion
        #endregion

        #region Direct3D10 Fields
        private Device10 device10;

        #region Graphic Setting
        #endregion
        #endregion

        private Matrix worldMatrix;

        #region Properties
        /// <summary>
        /// Getter for settings.
        /// </summary>
        public Setting Setting
        {
            get
            {
                switch (d3dVersion)
                {
                    case D3DVersion.Direct3D9:
                        return setting9;
                    case D3DVersion.Direct3D10:
                        return setting10;
                    default:
                        return setting9;
                }
            }
        }

        /// <summary>
        /// Getter for device.
        /// </summary>
        public Object Device
        {
            get
            {
                switch (d3dVersion)
                {
                    case D3DVersion.Direct3D9:
                        return device9;
                    case D3DVersion.Direct3D10:
                        return device10;
                    default:
                        return device9;
                }
            }
        }

        /// <summary>
        /// Returns Direct3D9 device.
        /// </summary>
        public Device9 Device9
        {
            get { return device9; }
        }

        /// <summary>
        /// Returns Direct3D9 interface.
        /// </summary>
        public D3D9.Direct3D Direct3D9
        {
            get { return d3d9; }
        }

        /// <summary>
        /// Returns Direct3D10 device.
        /// </summary>
        public Device10 Device10
        {
            get { return device10; }
        }

        /// <summary>
        /// Returns Direct3D9 setting.
        /// </summary>
        public Setting9 Setting9
        {
            get { return setting9; }
        }

        /// <summary>
        /// Returns Direct3D10 setting.
        /// </summary>
        public Setting10 Setting10
        {
            get { return setting10; }
        }

        /// <summary>
        /// Returns Camera class.
        /// </summary>
        public Camera Camera
        {
            get { return camera; }
        }

        /// <summary>
        /// Returns World matrix.
        /// </summary>
        public Matrix WorldMatrix
        {
            get { return worldMatrix; }
        }

        /// <summary>
        /// Progress of buffers initialization.
        /// </summary>
        public float Progress
        {
            get { return progress; }
            set { progress = value; }
        }
        #endregion
        #endregion

        /// <summary>
        /// Simple Constructor
        /// </summary>
        public GlobalManager()
        {
            setting9 = new Setting9();
            setting10 = new Setting10();
            camera = new Camera();
        }

        #region Public Methods
        /// <summary>
        /// This method setup the attributes.
        /// </summary>
        /// <param name="bgColor">Background color of rendering window.</param>
        /// <param name="meshCount">Number of meshes.</param>
        /// <param name="sameConnectivity">Is mesh connected with it selves.</param>
        /// <param name="showColorAttribute">Determines if the color attribute will be shown.</param>
        /// <param name="colorAttributeName">Name of the color attribute.</param>
        /// <param name="showVectorAttribute">Determines if the vector attribute will be shown.</param>
        /// <param name="vectorAttributeName">Determines if the vector attribute will be shown.</param>
        public void Setup(ColorRGBA bgColor, int meshCount, bool sameConnectivity, bool showColorAttribute, string colorAttributeName,
            bool showVectorAttribute, string vectorAttributeName)
        {
            Setting.CurrentMesh = 0;
            this.backColor = new Color4(bgColor.AlphaF, bgColor.RedF, bgColor.GreenF, bgColor.BlueF);
            Setting.MeshCount = meshCount;
            this.sameCon = sameConnectivity;
            this.showColor = showColorAttribute;
            this.colorAttrName = colorAttributeName;
            this.showVector = showVectorAttribute;
            this.vectorAttrName = vectorAttributeName;
            this.progress = 0.0f;
            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    if (sameCon)
                        objects3D9 = new Object3D9[1];
                    else
                        objects3D9 = new Object3D9[meshCount];
                    //TODO: objects3D10 = null;
                    break;
            }
        }

        /// <summary>
        /// Initialize window and save the handle of the rendering control.
        /// </summary>
        /// <param name="handle">Handle to control, witch will be rendered to.</param>
        /// <param name="clientSize">Client size of the control.</param>
        public void Initialize(IntPtr handle, Size clientSize)
        {
            this.hWnd = handle;
            this.clientSize = clientSize;

            worldMatrix = Matrix.Translation(new Vector3(0.0f, 0.0f, 0.0f));

            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    InitDirect3D9();
                    this.device10 = null; //Only one device is used.
                    break;
                case D3DVersion.Direct3D10:
                    InitDirect3D10();
                    this.device9 = null; //Only one device is used.
                    break;
            }
        }

        /// <summary>
        /// Initialize buffers for all meshes.
        /// </summary>
        public void InitializeBuffers()
        {
            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    for(int i = 0; i < objects3D9.Length; i++)
                    {
                        objects3D9[i].InitBuffersTriangle();
                        Progress = (float)i / Setting.MeshCount; 
                    }
                    Progress = 1.0f;
                    break;
            }
        }

        /// <summary>
        /// Save all points, primitives, colors etc. for the mesh and prepare for index and vertex buffer initialization.
        /// </summary>
        /// <param name="index">Index of mesh.</param>
        /// <param name="mesh">Mesh witch will be used as source.</param>
        public void SetMesh(int index, UnstrGrid mesh)
        {
            //SetTransforms(mesh);
            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    if (index > objects3D9.Length)
                        throw new MveException(this, "Index is out of bounds.");

                    objects3D9[index] = new Object3D9(showColor, showVector);
                    objects3D9[index].SetMesh(mesh);
                    break;
            }
        }

        /// <summary>
        /// Calculate optimal size and position of mesh
        /// </summary>
        /// <param name="mesh">Mesh witch transformations will be calculated</param>
        public void SetTransforms(UnstrGrid mesh)
        {
            Matrix toCenter, scale;
            scale = Matrix.Identity;
            Zcu.Mve.Visualization.BoundingBox bb = mesh.BoundingBox;
            toCenter = Matrix.Identity * Matrix.Translation((float)-bb.MidPoint[0], (float)-bb.MidPoint[1], (float)-bb.MidPoint[2]);
            float h = (float)(bb.MaxPoint[1] - bb.MinPoint[1]);
            float w = (float)(bb.MaxPoint[0] - bb.MinPoint[0]);
            float ar = ((float)(this.clientSize.Width) / this.clientSize.Height);
            float k1 = 1 / h;
            float k2 = ar / w;
            if (h == 0) scale *= Matrix.Scaling(k2, k2, k2);
            else
                if (w == 0) scale *= Matrix.Scaling(k1, k1, k1);
                else
                    if (k1 > k2) scale *= Matrix.Scaling(k1, k1, k1);
                    else
                        scale *= Matrix.Scaling(0.5f * k2, 0.5f * k2, 0.5f * k2);
            worldMatrix = toCenter * scale;
        }

        /// <summary>
        /// Public method for Update input.
        /// </summary>
        public void Update()
        {
            if (acquired)
            {
                UpdateKeyboard();
                UpdateMouse();
            }

        }

        /// <summary>
        /// Public method for Update keyboard. (Called in Update()).
        /// </summary>
        public void UpdateKeyboard()
        {
            if (keyboard.Acquire().IsFailure)
                return;

            if (keyboard.Poll().IsFailure)
                return;

            keyState = keyboard.GetCurrentState();
            if (Result.Last.IsFailure)
                return;

            #region Key state tests
            if (KeypressTest(Key.Y))
            {
                switch (Setting9.Fillmode)
                {
                case D3D9.FillMode.Point:
                    Setting9.Fillmode = D3D9.FillMode.Solid;
                    break;
                case D3D9.FillMode.Solid:
                    Setting9.Fillmode = D3D9.FillMode.Wireframe;
                    break;
                case D3D9.FillMode.Wireframe:
                    Setting9.Fillmode = D3D9.FillMode.Point;
                    break;
                }
            }

            if (KeypressTest(Key.C))
            {
                switch (Setting9.Culling)
                {
                    case D3D9.Cull.Clockwise:
                        Setting9.Culling = D3D9.Cull.Counterclockwise;
                        break;
                    case D3D9.Cull.Counterclockwise:
                        Setting9.Culling = D3D9.Cull.None;
                        break;
                    case D3D9.Cull.None:
                        Setting9.Culling = D3D9.Cull.Clockwise;
                        break;
                }
            }

            if (KeypressTest(Key.N))
            {
                Setting9.ShowNormals = !Setting9.ShowNormals;
            }

            if (KeypressTest(Key.X))
            {
                Setting9.Lighting = !Setting9.Lighting;
            }

            if (KeypressTest(Key.RightArrow))
            {
                Setting.CurrentMesh++;
            }

            if (KeypressTest(Key.LeftArrow))
            {
                Setting.CurrentMesh--;
            }

            if (KeypressTest(Key.Escape))
            {
                keyboard.Unacquire();
            }

            if (KeypressTest(Key.Home))
            {
                camera.SavePosition("camera_position.bin");
            }

            if (KeypressTest(Key.End))
            {
                camera.LoadPosition("camera_position.bin");
            }

            if (KeypressTest(Key.F11))
            {
                CaptureScreenshot9("screenshot", numScreenShots++);
            }
            #endregion

            oldKeyState = keyState;
        }

        /// <summary>
        /// Public method for Update mouse. (Called in Update()).
        /// </summary>
        public void UpdateMouse()
        {
            if (mouse.Acquire().IsFailure)
                return;

            if (mouse.Poll().IsFailure)
                return;

            mouseState = mouse.GetCurrentState();
            if (Result.Last.IsFailure)
                return;

            #region Mouse state tests
            // Mouse left button
            if (mouseState.IsPressed((int)MouseObject.Button1))
            {
                camera.Rotate((float)mouseState.X, (float)mouseState.Y, 0.0f);
            }

            //Mouse right button
            if (mouseState.IsPressed((int)MouseObject.Button2))
            {

            }

            //Mouse middle button
            if (mouseState.IsPressed((int)MouseObject.Button3))
            {
                if (keyState.IsPressed(Key.LeftShift) || keyState.IsPressed(Key.RightShift))
                {
                    camera.Move((float)mouseState.X, 0.0f, (float)mouseState.Y);
                }
                else
                {
                    camera.Move((float)mouseState.X, (float)mouseState.Y, 0.0f);
                }
                
            }

            if (mouseState.Z != 0)
            {
                camera.Zoom(-(float)mouseState.Z / 60);
            }

            //Keyboard test
            if (keyState.IsPressed(Key.L))
            {
                //TODO: Hodit do Properties (Setting)
                lightAzimuth -= (float)mouseState.X / 100f;
                lightElevation -= (float)mouseState.Y / 100f;
                globalLight.Direction = new Vector3((float)(Math.Cos(-lightAzimuth) * Math.Cos(lightElevation)),
                                                    (float)(Math.Sin(lightElevation)),
                                                    (float)(Math.Sin(-lightAzimuth) * Math.Cos(lightElevation)));
                device9.SetLight(0, globalLight);
                device9.EnableLight(0, true);
            }
            #endregion
        }

        /// <summary>
        /// Keypress Test method check if the key is pressed or not.
        /// </summary>
        /// <param name="theKey">What key will be tested</param>
        /// <returns>Returns true if the key is pressed</returns>
        private bool KeypressTest(Key theKey)
        {
            if (keyState.IsPressed(theKey) && oldKeyState.IsReleased(theKey))
                return true;
            return false;
        }

        /// <summary>
        /// Mousepress Test method check if the mouse button is pressed or not.
        /// </summary>
        /// <param name="button">What button will be tested</param>
        /// <returns>Returns true if the key is pressed</returns>
        private bool MousepressTest(MouseObject button)
        {
            if (mouseState.IsPressed((int)button) && oldMouseState.IsReleased((int)button))
                return true;
            return false;
        }

        /// <summary>
        /// Public method for Render event.
        /// </summary>
        public void Render()
        {
            switch (d3dVersion)
            {
                case D3DVersion.Direct3D9:
                    Render9();
                    this.device10 = null; //Only one device is used.
                    break;
                case D3DVersion.Direct3D10:
                    Render10();
                    this.device9 = null; //Only one device is used.
                    break;
            }
        }

        /// <summary>
        /// Resets the device if resizing or lost device.
        /// </summary>
        public void Reset()
        {
            Reset(this.clientSize);
        }

        /// <summary>
        /// Resets the device if resizing or lost device.
        /// </summary>
        public void Reset(Size clientSize)
        {
            this.clientSize = clientSize;

            if (device9 != null)
            {
                SetPresentParameters9();
                device9.Reset(presentParameters);
                SetRenderStates9();
                SetGlobalLight9();
                SetView9();
            }
            if (device10 != null && !device10.Disposed)
            {
                device10.Dispose();
                device10 = null;
            }
        }

        /// <summary>
        /// Sets the Direct3D version to be used.
        /// </summary>
        /// <param name="d3dVersion"></param>
        public void SetD3DVersion(D3DVersion d3dVersion)
        {
            this.d3dVersion = d3dVersion;
        }

        /// <summary>
        /// Dispose all resources.
        /// </summary>
        public void Dispose()
        {
            foreach (Object3D9 obj in objects3D9)
            {
                obj.Dispose();
            }
            if (d3d9 != null && !d3d9.Disposed)
            {
                d3d9.Dispose();
                d3d9 = null;
            }
            if (device9 != null && !device9.Disposed)
            {
                device9.Dispose();
                device9 = null;
            }
            if (device10 != null && !device10.Disposed)
            {
                device10.Dispose();
                device10 = null;
            }
            InputManager.Instance.Dispose();
        }
        #endregion

        #region Direct3D9
        /// <summary>
        /// Initialize Direct3D9 device.
        /// </summary>
        private void InitDirect3D9()
        {
            SetPresentParameters9();
            d3d9 = new D3D9.Direct3D();
            device9 = new D3D9.Device(d3d9, 0, D3D9.DeviceType.Hardware, this.hWnd, D3D9.CreateFlags.HardwareVertexProcessing, presentParameters);
            SetRenderStates9();
            SetGlobalLight9();
            SetView9();
        }

        /// <summary>
        /// Sets present parameters for Direct3D9 device.
        /// </summary>
        private void SetPresentParameters9()
        {
            presentParameters = new D3D9.PresentParameters()
            {
                SwapEffect = D3D9.SwapEffect.Discard,
                EnableAutoDepthStencil = true,
                AutoDepthStencilFormat = D3D9.Format.D16,
                BackBufferFormat = D3D9.Format.X8R8G8B8,
                BackBufferHeight = this.clientSize.Height,
                BackBufferWidth = this.clientSize.Width,
                Multisample = setting9.MultisampleType,
            };
        }

        /// <summary>
        /// Sets global render states for Direct3D9 device.
        /// </summary>
        private void SetRenderStates9()
        {
            device9.SetRenderState(D3D9.RenderState.FillMode, setting9.Fillmode);
            device9.SetRenderState(D3D9.RenderState.CullMode, setting9.Culling);
            device9.SetRenderState(D3D9.RenderState.ZEnable, true);
            device9.SetRenderState(D3D9.RenderState.Lighting, setting9.Lighting);
            device9.SetRenderState(D3D9.RenderState.Ambient, setting9.Ambient.ToArgb());
            device9.SetRenderState(D3D9.RenderState.ShadeMode, setting9.Shademode);
        }

        /// <summary>
        /// Sets the global light for Direct3D9 device.
        /// </summary>
        private void SetGlobalLight9()
        {
            globalLight = new D3D9.Light();
            globalLight.Diffuse = Color.DarkGray;
            globalLight.Type = D3D9.LightType.Directional;
            globalLight.Direction = new Vector3((float)(Math.Cos(-lightAzimuth) * Math.Cos(lightElevation)),
                                                    (float)(Math.Sin(lightElevation)),
                                                    (float)(Math.Sin(-lightAzimuth) * Math.Cos(lightElevation)));
            globalLight.Specular = Color.DarkGray;
            globalLight.Range = 50.0f;

            device9.SetLight(0, globalLight);
            device9.EnableLight(0, true);
        }

        /// <summary>
        /// Sets the camera position and perspective for Direct3D9 device.
        /// </summary>
        private void SetView9()
        {
            camera.LookAt(new Vector3(0.0f, 0.0f, 20.0f) ,new Vector3(0.0f, 0.0f, 0.0f), Vector3.UnitY);
            camera.Perspective(45, (float)this.clientSize.Width / this.clientSize.Height, 0.01f, 1000.0f);

            device9.SetTransform(D3D9.TransformState.View, camera.ViewMatrix);
            device9.SetTransform(D3D9.TransformState.Projection, camera.ProjectionMatrix);
        }

        /// <summary>
        /// Render mesh with Direct3D9 device.
        /// </summary>
        private void Render9()
        {
            try
            {
                device9.SetTransform(D3D9.TransformState.View, camera.ViewMatrix);
                device9.SetTransform(D3D9.TransformState.Projection, camera.ProjectionMatrix);
                device9.SetTransform(D3D9.TransformState.World, worldMatrix);

                device9.Clear(D3D9.ClearFlags.Target | D3D9.ClearFlags.ZBuffer, backColor, 1.0f, 0);
                device9.BeginScene();

                objects3D9[Setting.CurrentMesh].DrawMesh();

                if (setting9.ShowNormals)
                    objects3D9[Setting.CurrentMesh].DrawNormals();

                device9.EndScene();
                device9.Present();
            }
            catch (D3D9.Direct3D9Exception ex)
            {
                System.Windows.Forms.MessageBox.Show(ex.ToString());
            }
        }

        /// <summary>
        /// Capture screenshot of window frame
        /// </summary>
        /// <param name="name"></param>
        /// <param name="number"></param>
        private void CaptureScreenshot9(String name, int number)
        {
            D3D9.Surface backbuffer = device9.GetBackBuffer(0, 0);
            D3D9.Surface.ToFile(backbuffer,  name + "_" + number + ".bmp", D3D9.ImageFileFormat.Bmp);
            backbuffer.Dispose();
        }
        #endregion

        #region Direct3D10
        /// <summary>
        /// Initialize Direct3D10 device.
        /// </summary>
        private void InitDirect3D10()
        {

        }
        /// <summary>
        /// Render mesh with Direct3D10 device.
        /// </summary>
        private void Render10()
        {

        }
        #endregion

        #region Thread-safe Singleton Pattern
        /// <summary>
        /// Thread-safe singleton instance.
        /// </summary>
        public static GlobalManager Instance
        {
            get
            {
                return GMNested.instance;
            }
        }

        /// <summary>
        /// Thread-safe singleton protection class.
        /// </summary>
        class GMNested
        {
            /// <summary>
            /// Static constructor.
            /// </summary>
            static GMNested()
            {
            }

            internal static readonly GlobalManager instance = new GlobalManager();
        }
        #endregion
    }

    
}
