using System;
using System.IO;
using System.ComponentModel;
using Framework;

namespace MeshBrowser.SlimDXRenderer
{
	/// <summary>
	/// Main class of MeshLoader. Loads and saves configuration file and runs file readers.
	/// </summary>	
	public class MeshLoader
	{
		/// <summary>
		/// Stores type of input file.
		/// </summary>
		internal byte typeOfFile = 0;

		/// <summary>
		/// Stores name of input file.
		/// </summary>
		internal string nameOfFile = null;

        ProgressInfo progress;

        public ProgressInfo Progress
        {
            get { return progress; }
            set { progress = value; }
        }

        
		public string FileName
		{
			get { return nameOfFile; }
			set
			{
				byte type;

				if (!File.Exists(value))          // Test if the file exists.
				{
					Console.WriteLine("Invalid file name!");					
				}

				if ((type = this.tellFileType(value.ToLower())) == 0)
				{                                              // Test the extension of file - 0 is code for unsupported type.
					Console.WriteLine("Unknown file format!");
				}

				this.nameOfFile = value;
				// Change meshLoader file name property.
				this.typeOfFile = type;             // Change meshLoader file type property.

				this.newFileSelected = true;    // ... set flag that cause generating new triangle mesh.

			}
		}

		/// <summary>
		/// Indicates that new file (filename) is selected.
		/// </summary>
		internal bool newFileSelected = true;

		/// <summary>
		/// Reference of DataSet class that contains mesh data.
		/// </summary>
		private DataSet dataSet = null;

		/// <summary>
		/// Mesh data that is on output port of module.
		/// </summary>
        private TriangleMesh mesh = null;

        public TriangleMesh Mesh
        {
            get { return mesh; }
            set { mesh = value; }
        }		

		/// <summary>
		/// Entry point of module. Reads input file and sets data on output port.
		/// </summary>
		public void Execute()
		{
			
            if ((this.newFileSelected) && this.typeOfFile != 0)
            {
                this.newFileSelected = false;

                try
                {
                    if (!System.IO.File.Exists(this.nameOfFile))
                    {
                        throw new FileNotFoundException("");
                    }

                    int startTics = Environment.TickCount;
                    this.ReadFile();
                    int endTics = Environment.TickCount;

                    Console.WriteLine("File loaded in {0} ms", (endTics - startTics));
                }
                catch (FileNotFoundException)
                {
                    this.newFileSelected = true;

                    throw new Exception("Input file not found.");
                }

                //catch (System.Exception)
                //{
                //    this.newFileSelected = true;
                //    throw new Exception("Fatal error in meshLoader - uncaught System exception.");
                //}

                this.mesh = this.dataSet.GetMesh();        // All is OK - loaded mesh is on output port.                
            }            
		}


		// LVA 06/07/06 - removed in order to use the srandard setup dialog
		/*
		/// <summary>
		/// Creates MeshLoaderSetup, where user can sets input file.
		/// </summary>
		/// <returns>Module setup.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new MeshLoaderSetup(this, this.nameOfFile);
		}*/

		/// <summary>
		/// Configuration is saved to XML map file.
		/// </summary>
		/// <param name="config">Element of XML that will be saved in map file.</param>
		public void WriteConfig(System.Xml.XmlElement config)
		{
			System.Xml.XmlElement xmlEl;

			xmlEl = config.OwnerDocument.CreateElement("Output_file");
			xmlEl.SetAttribute("path", this.nameOfFile);
			xmlEl.SetAttribute("type", this.typeOfFile.ToString());
			config.AppendChild(xmlEl);
		}

		/// <summary>
		/// Analyze type of selected file based on file extension.
		/// </summary>
		/// <param name="path">Full filename (full path and filename).</param>
		/// <returns>Return type of selected input file - 1 - TRI format,
		///                                               2 - STL format,
		///                                               3 - PLY format,
		///                                               4 - CMX format,
		///												  5 - WRL format,
		///                                               0 - ??? format.</returns>
		private byte tellFileType(string path)
		{
			if (path.EndsWith(".tri"))
			{
				return 1;
			}
			else if (path.EndsWith(".stl"))
			{
				return 2;
			}
			else if (path.EndsWith(".ply"))
			{
				return 3;
			}
			else if (path.EndsWith(".cmx"))
			{
				return 4;
			}
			else if (path.EndsWith(".wrl"))
			{
				return 5;
			}
			else
			{
				return 0;
			}
		}

		/// <summary>
		/// Call methods for read input file based on type of file.
		/// </summary>
		private void ReadFile() 
		{
			switch (this.typeOfFile) 
			{
				case 1: this.ReadTriFile();
					break;
				case 2: this.ReadStlFile();
					break;
				case 3: this.ReadPlyFile();
					break;
				case 4: this.ReadCmxFile();
					break;
				case 5: this.ReadWrlFile();
					break;

			}
		}

		/// <summary>
		/// Creates reader object of TRI format.
		/// </summary>
		private void ReadTriFile()
		{
			ReaderTri reader = new ReaderTri();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile, progress);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of WRL format.
		/// </summary>
		private void ReadWrlFile()
		{
			ReaderWrl reader = new ReaderWrl();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			// Reads file.
		}

		/// <summary>
		/// Analyze input file with STL extension and decide if it's binary or text.
		/// </summary>
		private void ReadStlFile()
		{
            using (StreamReader streamReader = new StreamReader(this.nameOfFile))
            {

                string line = streamReader.ReadLine().ToLower();

                if (line.StartsWith("solid"))
                {
                    this.ReadAsciiStlFile();
                }
                else
                {
                    this.ReadBinStlFile();
                }
            }
		}

		/// <summary>
		/// Creates reader object of text STL format.
		/// </summary>
		private void ReadAsciiStlFile()
		{
			ReaderStl reader = new ReaderStl();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of binary STL format.
		/// </summary>
		private void ReadBinStlFile()
		{
			ReaderBinStl reader = new ReaderBinStl();      // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile, progress);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of PLY format.
		/// </summary>
		private void ReadPlyFile()
		{
			ReaderPly reader = new ReaderPly();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile, progress);
                                                           // Reads file.
		}

		/// <summary>
		/// Creates reader object of CMX format.
		/// </summary>
		private void ReadCmxFile()
		{
			ReaderCmx reader = new ReaderCmx();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
                                                           // Reads file.
		}
	}
}
