﻿using System;
using System.Drawing;
using System.ComponentModel;
using SlimDX;
using Matrix = SlimDX.Matrix;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using SceneMgr = SlimDXRenderSystem.Managers.SceneManager;
using InputMgr = SlimDXRenderSystem.Managers.InputManager;
using TimeMgr = SlimDXRenderSystem.Managers.TimeManager;
using System.Threading;
using Framework;
using System.Windows;
using System.Reflection;

namespace MeshBrowser.SlimDXRenderer
{

    public enum PickType
    {
        NONE,
        POINTS,
        PRIMITIVES
    }

    /// <summary>
    /// This class manage all module stuff like objects, cameras, lights etc.
    /// </summary>
    public sealed class ModuleManager : SlimDXRenderSystem.Core.Singleton<ModuleManager>
    {
        #region Fields
        PickType pick = PickType.NONE;
        private Color4 backColor;
        private int numScreenShots = 0;
        private float progress = 0.0f;
        private bool sameCon;
        private bool showColor;
        private bool showEdgeColor;
        private bool showVector;
        private int currentMesh;
        private int meshCount;

        private double mouseX;
        private double mouseY;

        private int pickedPointId;
        private int pickedPrimId;

        private System.Windows.Size clientSize;

        private SlimDXRenderSystem.Core.Group meshGroup;
        private SlimDXRenderSystem.Core.Group cursorGroup;

        
        private Object3D[] objects;
        private Object3D cursor;

        private RenderSetting rendersetting;
        private InputSetting inputsetting;

        private Action<int, double, double, double> vertexPickAction;
        private Action<int, int, int, int> primitivePickAction;

        

        #region Properties
        public PickType Pick
        {
            get { return pick; }
            set { pick = value; }
        }

        public Action<int, double, double, double> VertexPickAction
        {
            get { return vertexPickAction; }
            set { vertexPickAction = value; }
        }

        public Action<int, int, int, int> PrimitivePickAction
        {
            get { return primitivePickAction; }
            set { primitivePickAction = value; }
        }

        public int NumScreenShots
        {
            get { return numScreenShots; }
            set { numScreenShots = value; }
        }

        /// <summary>
        /// Progress of buffers initialization.
        /// </summary>
        public float Progress
        {
            get { return progress; }
            set { progress = value; }
        }

        public RenderSetting Rendersetting
        {
            get { return rendersetting; }
        }

        public InputSetting Inputsetting
        {
            get { return inputsetting; }
        }

        public int CurrentMesh
        {
            get { return currentMesh; }
            set
            {
                CurrentObject.Object.Visible = false;
                if (value >= meshCount)
                    currentMesh = 0;
                else if (value < 0)
                    currentMesh = meshCount - 1;
                else
                    currentMesh = value;
                CurrentObject.Object.Visible = true;
            }
        }

        public int MeshCount
        {
            get { return meshCount; }
        }

        public SlimDXRenderSystem.Core.Group MeshGroup
        {
            get { return meshGroup; }
        }

        public SlimDXRenderSystem.Core.Group CursorGroup
        {
            get { return cursorGroup; }
            set { cursorGroup = value; }
        }

        public Object3D CurrentObject
        {
            get { return objects[currentMesh]; }
        }

        public System.Windows.Size ClientSize
        {
            get { return clientSize; }
            set { clientSize = value; }
        }

        public double MouseX
        {
            get { return mouseX; }
            set { mouseX = value; }
        }

        public double MouseY
        {
            get { return mouseY; }
            set { mouseY = value; }
        }

        #endregion
        #endregion

        #region Public Methods
        /// <summary>
        /// This method setup the attributes.
        /// </summary>
        /// <param name="bgColor">Background color of rendering window.</param>
        /// <param name="meshCount">Number of meshes.</param>
        /// <param name="sameConnectivity">Is mesh connected with it selves.</param>
        /// <param name="showColorAttribute">Determines if the color attribute will be shown.</param>
        /// <param name="colorAttributeName">Name of the color attribute.</param>
        /// <param name="showVectorAttribute">Determines if the vector attribute will be shown.</param>
        /// <param name="vectorAttributeName">Determines if the vector attribute will be shown.</param>
        public void Setup(ColorRGBA bgColor, int meshCount, bool sameConnectivity, bool showColorAttribute,
            bool showVectorAttribute, bool showEdgeColorAttribute)
        {
            this.backColor = new Color4(bgColor.AlphaF, bgColor.RedF, bgColor.GreenF, bgColor.BlueF);
            this.sameCon = sameConnectivity;
            this.showColor = showColorAttribute;
            this.showVector = showVectorAttribute;
            this.showEdgeColor = showEdgeColorAttribute;
            this.progress = 0.0f;
            this.currentMesh = 0;
            this.meshCount = meshCount;

            objects = new Object3D[meshCount];

            inputsetting = new InputSetting();
            rendersetting = new RenderSetting();

            SceneMgr.Instance.CreateCamera("Perspective");
            SceneMgr.Instance.Camera.Perspective(45, (float)(clientSize.Width / clientSize.Height), 0.1f, 100.0f);
            SceneMgr.Instance.Camera.LookAt(new Vector3(0.0f, 0.0f, 20.0f), new Vector3(0.0f, 0.0f, 0.0f), Vector3.UnitY);
            RenderMgr.Instance.RenderTextAction = WriteText;
        }

        /// <summary>
        /// Initialize buffers for all meshes.
        /// </summary>
        public void Initialize(bool keepPosition)
        {
            try
            {
                cursor.InitVertexBuffers();
                cursor.InitIndexBuffers();
                cursor.setVisible(false);                
                cursorGroup = SceneMgr.Instance.CreateGroup("Cursor", null);
                cursorGroup.AddObject(cursor.Object);
                //cursorGroup.Position = -Vector3.TransformCoordinate(new Vector3(0, 0, 0), SceneMgr.Instance.WorldMatrix);
             
                if (keepPosition)
                    meshGroup = SceneMgr.Instance.CreateGroup("Object", meshGroup);
                else
                    meshGroup = SceneMgr.Instance.CreateGroup("Object", null);

                for (int i = 0; i < objects.Length; i++)
                {
                    if (objects[i] != null)
                    {
                        objects[i].InitVertexBuffers();
                        if (sameCon && i != 0)
                            objects[i].Object.SetMeshSameIndexBuffer(objects[0].Object);
                        else
                            objects[i].InitIndexBuffers();

                        meshGroup.AddObject(objects[i].Object); // Add to group
                    }
                    Progress = (float)(i + 1) / objects.Length;
                }
                Progress = 1.0f;
                GC.Collect();
                objects[0].Object.Visible = true;

                RenderMgr.Instance.Renderer.BackgroundColor = backColor;
                
            }
            catch (ThreadAbortException e)
            {
                System.Windows.MessageBox.Show(e.StackTrace, e.Message);
                Thread.ResetAbort();
                throw new Exception("Thread needs to be aborted", e);
            }
        }

        /// <summary>
        /// Save all points, primitives, colors etc. for the mesh and prepare for index and vertex buffer initialization.
        /// </summary>
        /// <param name="index">Index of mesh.</param>
        /// <param name="mesh">Mesh witch will be used as source.</param>
        public void SetMesh(int index, TriangleMesh mesh)
        {
            if (index > objects.Length)
                throw new Exception("Index is out of bounds.");

            objects[index] = new Object3D("Mesh" + index.ToString(), showColor, showVector, showEdgeColor);
            objects[index].SetMesh(mesh);
            SceneMgr.Instance.CursorLight.Position = new Vector3(0, 0, 0);
        }

        /// <summary>
        /// Call the render method of rendermanager.
        /// </summary>
        public void Update()
        {
            
            
        }

        public void PickVertex(int index)
        {
            if (index < CurrentObject.Points.Length)
            {
                pickedPointId = index;
                if (vertexPickAction != null)
                    vertexPickAction(pickedPointId, ((Point3D)CurrentObject.Points[pickedPointId]).X, ((Point3D)CurrentObject.Points[pickedPointId]).Y, ((Point3D)CurrentObject.Points[pickedPointId]).Z);

            }
            else
                MessageBox.Show("Current mesh has only " + CurrentObject.Points.Length.ToString() + " vertices, cannot pick vertex no. " + index.ToString() + ".", "Index out of range", MessageBoxButton.OK, MessageBoxImage.Exclamation);
        }

        public void PickPrimitive(int index)
        {
            if (index < CurrentObject.Primitives.Length)
            {
                pickedPrimId = index;
                if (primitivePickAction != null)
                    primitivePickAction(pickedPrimId, ((Triangle)CurrentObject.Primitives[pickedPointId]).V1, ((Triangle)CurrentObject.Primitives[pickedPointId]).V2, ((Triangle)CurrentObject.Primitives[pickedPointId]).V3);

            }
            else
                MessageBox.Show("Current mesh has only " + CurrentObject.Points.Length.ToString() + " vertices, cannot pick vertex no. " + index.ToString() + ".", "Index out of range", MessageBoxButton.OK, MessageBoxImage.Exclamation);
        }

        public void CursorPicking()
        {
            if (pick == PickType.NONE)
                return;

            Matrix m = SceneMgr.Instance.WorldMatrix * Matrix.Translation(meshGroup.Position.X, meshGroup.Position.Y, meshGroup.Position.Z);
            Vector3 cursorPos = cursorGroup.Position;
            Point3D cursorPoint = new Point3D(cursorPos.X, cursorPos.Y, cursorPos.Z);
            Matrix mInv = m;
            mInv.Invert();
            Vector4 cursorPointInvTransformed = Vector3.Transform(cursorPos, mInv);
            Point3D cursorPoint3DInv = new Point3D(cursorPointInvTransformed.X, cursorPointInvTransformed.Y, cursorPointInvTransformed.Z);

            double minDist = double.MaxValue;
            pickedPointId = 0;

            switch (pick)
            {
                case PickType.POINTS:
                    PickVertex(CurrentObject.GetPointNearestTo(cursorPoint3DInv));                 
                    break;
                case PickType.PRIMITIVES:
                    int picked = 0;
                    for (int i = 0; i < CurrentObject.Primitives.Length; i++)
                    {
                        Point3D p = new Point3D();
                        //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                        {
                            Triangle t = (Triangle)CurrentObject.Primitives[i];
                            Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                            Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                            Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                            p = (p1 + p2 + p3) * (1 / 3.0);
                        }
                        
                        /*Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        p.X = v.X;
                        p.Y = v.Y;
                        p.Z = v.Z;*/

                        Point3D dif = p - cursorPoint3DInv;
                        double distance = dif.dotProduct(dif);

                        if (distance < minDist)
                        {
                            picked = i;
                            minDist = distance;
                        }
                    }
                    PickPrimitive(picked);
                    break;
            }
        }

        public void Picking()
        {
            if (pick == PickType.NONE)
                return;

            Matrix m = SceneMgr.Instance.WorldMatrix * Matrix.Translation(meshGroup.Position.X, meshGroup.Position.Y, meshGroup.Position.Z)
                * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix;
            int picked = 0;
            switch (pick)
            {
                case PickType.POINTS:
                    double minDist = double.MaxValue;
                    
                    for (int i = 0; i < CurrentObject.Points.Length; i++)
                    {
                        Point3D p = (Point3D)CurrentObject.Points[i];
                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        if (v.W <= 0)
                            continue;

                        int x = (int)(v.X * RenderMgr.Instance.RenderWindow.ActualWidth / 2 / (v.W) + RenderMgr.Instance.RenderWindow.ActualWidth / 2);
                        int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.ActualHeight / 2 / (v.W) + RenderMgr.Instance.RenderWindow.ActualHeight / 2);

                        if (x > RenderMgr.Instance.RenderWindow.ActualWidth || y > RenderMgr.Instance.RenderWindow.ActualHeight || x < 0 || y < 0)
                            continue;

                        double distance = (x - mouseX) * (x - mouseX) + (y - mouseY) * (y - mouseY);

                        if (distance < minDist)
                        {
                            picked = i;
                            minDist = distance;
                        }
                        
                    }
                    PickVertex(picked);
                    break;
                case PickType.PRIMITIVES:
                    minDist = double.MaxValue;
                    pickedPrimId = 0;

                    for (int i = 0; i < CurrentObject.Primitives.Length; i++)
                    {
                        Point3D p = new Point3D();
                        //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                        {
                            Triangle t = (Triangle)CurrentObject.Primitives[i];
                            Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                            Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                            Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                            p = (p1 + p2 + p3) * (1 / 3.0);
                        }
                        /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                        {
                            PolyLine l = (PolyLine)CurrentObject.Primitives[i];
                            p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                        }
                        else if (CurrentObject.Primitives.ElemType == typeof(Line))
                        {
                            Line l = (Line)CurrentObject.Primitives[i];
                            p = (Point3D)CurrentObject.Points[l.V1];
                        }*/

                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        if (v.W <= 0)
                            continue;

                        int x = (int)(v.X * RenderMgr.Instance.RenderWindow.ActualWidth / 2 / (v.W) + RenderMgr.Instance.RenderWindow.ActualWidth / 2);
                        int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.ActualHeight / 2 / (v.W) + RenderMgr.Instance.RenderWindow.ActualHeight / 2);

                        if (x > RenderMgr.Instance.RenderWindow.ActualWidth || y > RenderMgr.Instance.RenderWindow.ActualHeight || x < 0 || y < 0)
                            continue;

                        double distance = (x - mouseX) * (x - mouseX) + (y - mouseY) * (y - mouseY);

                        if (distance < minDist)
                        {
                            picked = i;
                            minDist = distance;
                        }
                    }
                    PickPrimitive(picked);
                    break;
            }
        }

        public void WriteText()
        {
            //Display Indices
            Matrix m = SceneMgr.Instance.WorldMatrix * Matrix.Translation(meshGroup.Position.X, meshGroup.Position.Y, meshGroup.Position.Z)
                * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix;

            if (rendersetting.DisplayIndices)
            {
                for (int i = Math.Max(0, rendersetting.MinIndex); i < Math.Min(CurrentObject.Points.Length, rendersetting.MaxIndex); i++)
                {
                    Point3D p = (Point3D)CurrentObject.Points[i];
                    Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                    Vector4 v = Vector3.Transform(v1, m);
                    RenderMgr.Instance.DrawText(v.X / v.W, v.Y/v.W, i.ToString());
                }
            }

            if (rendersetting.DisplayTriangleIndices)
            {
                for (int i = Math.Max(0, rendersetting.MinIndex); i < Math.Min(CurrentObject.Primitives.Length, rendersetting.MaxIndex); i++)
                {
                Point3D p = new Point3D();
                //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                {
                    Triangle t = (Triangle)CurrentObject.Primitives[i];
                    Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                    Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                    Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                    p = (p1 + p2 + p3) * (1 / 3.0);
                }
                /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                {
                    PolyLine l = (PolyLine)CurrentObject.Primitives[i];
                    p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                }
                else if (CurrentObject.Primitives.ElemType == typeof(Line))
                {
                    Line l = (Line)CurrentObject.Primitives[i];
                    p = (Point3D)CurrentObject.Points[l.V1];
                }*/
                Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                Vector4 v = Vector3.Transform(v1, m);

                RenderMgr.Instance.DrawText(v.X/v.W, v.Y/v.W, i.ToString());
                }
            }

            if (!rendersetting.DisplayIndices && !rendersetting.DisplayTriangleIndices)
            {
                switch (pick)
                {
                    case PickType.POINTS:
                        Point3D p = (Point3D)CurrentObject.Points[pickedPointId];
                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        RenderMgr.Instance.DrawPoint(new Vector3(v.X / v.W, v.Y / v.W, 0));
                        RenderMgr.Instance.DrawText(v.X/v.W, v.Y/v.W, pickedPointId.ToString());
                        break;

                    case PickType.PRIMITIVES:
                        p = new Point3D();
                        //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                        {
                            Triangle t = (Triangle)CurrentObject.Primitives[pickedPrimId];
                            Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                            Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                            Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                            p = (p1 + p2 + p3) * (1 / 3.0);

                            v1 = new Vector3((float)p1.X, (float)p1.Y, (float)p1.Z);
                            v = Vector3.Transform(v1, m);

                            RenderMgr.Instance.DrawPoint(new Vector3(v.X/v.W, v.Y/v.W, 0));

                            v1 = new Vector3((float)p2.X, (float)p2.Y, (float)p2.Z);
                            v = Vector3.Transform(v1, m);

                            RenderMgr.Instance.DrawPoint(new Vector3(v.X / v.W, v.Y / v.W, 0));

                            v1 = new Vector3((float)p3.X, (float)p3.Y, (float)p3.Z);
                            v = Vector3.Transform(v1, m);

                            RenderMgr.Instance.DrawPoint(new Vector3(v.X / v.W, v.Y / v.W, 0));

                        }
                        /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                        {
                            PolyLine l = (PolyLine)CurrentObject.Primitives[pickedPrimId];
                            p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                        }
                        else if (CurrentObject.Primitives.ElemType == typeof(Line))
                        {
                            Line l = (Line)CurrentObject.Primitives[pickedPrimId];
                            p = (Point3D)CurrentObject.Points[l.V1];
                        }*/

                        v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        v = Vector3.Transform(v1, m);

                        RenderMgr.Instance.DrawText(v.X/v.W, v.Y/v.W, pickedPrimId.ToString());
                        break;
                }
            }

            if (rendersetting.DisplayFPS)
                RenderMgr.Instance.DrawText(10, 10, "FPS: " + TimeMgr.Instance.FPS.ToString("F1"));            
            RenderMgr.Instance.DrawText(10, 30, "Current mesh: " + currentMesh.ToString());
        }

        /// <summary>
        /// Calculate optimal size and position of mesh
        /// </summary>
        /// <param name="mesh">Mesh witch transformations will be calculated</param>
        public void SetTransforms(TriangleMesh mesh)
        {

            var dpiXProperty = typeof(SystemParameters).GetProperty("DpiX", BindingFlags.NonPublic | BindingFlags.Static);

            var dpiX = (int)dpiXProperty.GetValue(null, null);

            var dpiratio = dpiX / 96.0;

            Matrix toCenter, scale;
            scale = Matrix.Identity;
            Framework.BoundingBox bb = mesh.BoundingBox;
            toCenter = Matrix.Identity * Matrix.Translation((float)-bb.MidPoint[0], (float)-bb.MidPoint[1], (float)-bb.MidPoint[2]);
            float w = (float)(bb.MaxPoint - bb.MinPoint).Abs();
            //float ar = (float)(clientSize.Width / clientSize.Height);
            float k2 = 1 / w;
            var c = 10;
            scale *= Matrix.Scaling((float)(k2*dpiratio*c) , (float)(k2*dpiratio*c) , (float)(k2*dpiratio*c) );
            SceneMgr.Instance.WorldMatrix = toCenter * scale;

            SceneMgr.Instance.SceneLights["Default"].Direction = new Vector3(1, 1, 1);
        }

        /// <summary>
        /// Dispose all resources.
        /// </summary>
        public void Dispose()
        {

        }
        #endregion

        internal void SetCursor(TriangleMesh mesh)
        {
            cursor = new Object3D("Cursor", false, false, false);
            cursor.SetMesh(mesh);             
        }

        public void Recenter()
        {
            MeshGroup.MoveTo(0,0,0);
            SceneMgr.Instance.Camera.ResetZoom();
            SceneMgr.Instance.Camera.LookAt(new Vector3(0.0f, 0.0f, 1.0f), new Vector3(0.0f, 0.0f, 0.0f), Vector3.UnitY);
        }
    }

    
}
