﻿using System;
using System.Drawing;
using SlimDX;
using SlimDXRenderSystem.Managers;
using Framework;
using System.Threading;

namespace MeshBrowser.SlimDXRenderer
{
    /// <summary>
    /// Default 3D Object that is derived by Direct3D objects..
    /// </summary>
    public class Object3D
    {
        #region Fields
        public static int MAXTRIANGLES = 3000000;
        private Framework.BoundingBox bb;
        private SlimDXRenderSystem.Core.Object3D obj;

        private bool showColor;
        private bool showEdgeColor;
        private bool showVector;

        private string name;

        #region UniformDataArrays
        protected Point3D[] points;
        protected Point3D[] normals;
        protected Triangle[] primitives;
        protected ColorRGBA[] colors;
        //protected UniformDataArray edgeColors;
        protected Point3D[] vectors;
        protected KDTree pointsTree;
        protected bool pointsTreeReady = false;
        #endregion

        #region Properties
        /// <summary>
        /// Show color attribute.
        /// </summary>
        public bool ShowColor
        {
            get { return showColor; }
            set { showColor = value; }
        }

        /// <summary>
        /// Show vector attribute.
        /// </summary>
        public bool ShowVector
        {
            get { return showVector; }
            set { showVector = value; }
        }        

        public Point3D[] Points
        {
            get { return points; }
        }

        public Triangle[] Primitives
        {
            get { return primitives; }
        }

        public SlimDXRenderSystem.Core.Object3D Object
        {
            get { return obj; }
        }

        #endregion
        #endregion

        /// <summary>
        /// Constructor.
        /// </summary>
        public Object3D()
        {
            this.name = "object";
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="showColor">Show color attribute.</param>
        /// <param name="showVector">Show vector attribute.</param>
        public Object3D(string name, bool showColor, bool showVector, bool showEdgeColor)
        {
            this.name = name;
            this.showColor = showColor;
            this.showEdgeColor = showEdgeColor;
            this.showVector = showVector;
        }

        /// <summary>
        /// Set mesh points, primitves, colors and vectors.
        /// </summary>
        /// <param name="mesh"></param>
        public void SetMesh(TriangleMesh mesh)
        {
            pointsTreeReady = false;
            this.bb = mesh.BoundingBox;
            this.points = mesh.Points;

            primitives = mesh.Triangles;

            if (showColor)
            {
                colors = mesh.Colors;
            }

            if (showVector)
            {                
                vectors = mesh.Vectors;
            }

            /*if (showEdgeColor)
            {
                if (colorEdgeAttrName == "")
                {
                    foreach (string attrId in mesh.GetCellAttrsIds("edges"))
                    {
                        UniformDataArray attr = mesh.GetCellAttrs("edges", attrId);
                        if (attr.ElemType == typeof(ColorRGBA))
                            edgeColors = attr;
                    }
                    if (edgeColors == null)
                        throw new Exception("ColorRGBA attribute not found.");
                }
                else
                    edgeColors = mesh.GetCellAttrs("edges", colorEdgeAttrName);
            }*/

            this.normals = mesh.Normals;
            /*else
                throw new MveException(this, "Normal attribute not found. Use the NormalComputer module to evaluate vertex normals");*/
            Thread KDBuilder = new Thread(new ThreadStart(BuildPointsTree));
            KDBuilder.Start();
        }

        private void BuildPointsTree()
        {
            this.pointsTree = new KDTree(this.points);
            pointsTreeReady = true;
        }

        public void InitVertexBuffers()
        {
            //if (primitives.ElemType == typeof(Triangle))
                InitVBTriangle();
            /*else if (primitives.ElemType == typeof(Line))
                InitVBLine();
            else if (primitives.ElemType == typeof(PolyLine))
                InitVBPolyLine();
            else
                throw new MveException(this, "Unexpected primitive type " + primitives.ElemType.ToString() + "!");*/
        }

        public void InitIndexBuffers()
        {
            //if (primitives.ElemType == typeof(Triangle))
                InitIBTriangle();
            /*else if (primitives.ElemType == typeof(Line))
                InitIBLine();
            else if (primitives.ElemType == typeof(PolyLine))
                InitIBPolyLine();
            else
                throw new MveException(this, "Unexpected primitive type " + primitives.ElemType.ToString() + "!");*/
        }

        /// <summary>
        /// Initialize index and vertex buffers for TriangleMesh type mesh
        /// </summary>
        public void InitVBTriangle()
        {
            int pCount = points.Length;

            Point3D diag = (Point3D)bb.MaxPoint - (Point3D)bb.MinPoint;
            double diagonal = Math.Sqrt(diag.dotProduct(diag));
            double multiplier = diagonal * 0.05;

            // Vertex Buffer
            Vector3[] vertices = new Vector3[showVector ? pCount * 3 : pCount * 2];
            Vector3[] vernormals = new Vector3[showVector ? pCount * 3 : pCount * 2];
            int[] vercolors = new int[showVector ? pCount * 3 : pCount * 2];
            for (int i = 0; i < pCount; i++)
            {
                Point3D point = (Point3D)points[i];
                Point3D normal;
                if (normals == null)
                    normal = new Point3D();
                else
                    normal = (Point3D)normals[i];

                vertices[i] = new Vector3((float)point.X, (float)point.Y, (float)point.Z);
                vernormals[i] = new Vector3((float)-normal.X, (float)-normal.Y, (float)-normal.Z);

                vertices[i + pCount] = new Vector3((float)(point.X + normal.X * multiplier), (float)(point.Y + normal.Y * multiplier), (float)(point.Z + normal.Z * multiplier));
                vernormals[i + pCount] = new Vector3((float)-normal.X, (float)-normal.Y, (float)-normal.Z);
                vercolors[i + pCount] = new ColorRGBA(1.0f, 0.0f, 0.0f).GetInt();

                if (ShowColor&&colors!=null)
                    vercolors[i] = ((ColorRGBA)colors[i]).GetInt();
                else
                    vercolors[i] = new ColorRGBA(0.5f, 0.5f, 0.5f).GetInt();

                if (ShowVector&&(vectors!=null))
                {
                    Point3D vector = vectors[i];
                    vertices[i + pCount] = new Vector3((float)(point.X + vector[0]), (float)(point.Y + vector[1]), (float)(point.Z + vector[2]));
                    vernormals[i + pCount] = new Vector3(1.0f, 0.0f, 0.0f);

                    if (ShowColor)
                        vercolors[i + pCount] = ((ColorRGBA)colors[i]).GetInt();
                    else
                        vercolors[i + pCount] = new ColorRGBA(0.5f, 0.5f, 0.5f).GetInt();
                }
            }
            obj = SceneManager.Instance.CreateObject(name);
            obj.SetMeshVertices(vertices, vernormals, vercolors, pCount);
        }

        internal void setVisible(bool v)
        {
            this.obj.Visible = v;
        }

        public void InitIBTriangle()
        {
            int pCount = points.Length;

            // main mesh indices
            int tCount = primitives.Length;
            int nIBuffers = (tCount) / MAXTRIANGLES;
            uint[] triangleIndices;
            for (int ib = 0; ib <= nIBuffers; ib++)
            {
                if (ib == nIBuffers)
                    triangleIndices = new uint[3*(tCount % MAXTRIANGLES)];
                else
                    triangleIndices = new uint[3*MAXTRIANGLES];
                for (uint i = 0; i < triangleIndices.Length / 3; i++)
                {
                    Triangle triangle = (Triangle)primitives[(int)i + (ib * MAXTRIANGLES)];
                    triangleIndices[3 * i] = (uint)triangle.V1;
                    triangleIndices[3 * i + 1] = (uint)triangle.V2;
                    triangleIndices[3 * i + 2] = (uint)triangle.V3;
                }
                obj.AddMeshIndices("main" + ib, SlimDXRenderSystem.MeshType.TriangleMesh, triangleIndices);
                obj.EnableDrawIndices("main" + ib);
            }
            // Normals indices
            triangleIndices = new uint[pCount * 2];
            for (uint i = 0; (i < pCount) && (i < uint.MaxValue); i++)
            {
                triangleIndices[2 * i] = (uint)i;
                triangleIndices[2 * i + 1] = (uint)(i + pCount);
            }
            obj.AddMeshIndices("normal", SlimDXRenderSystem.MeshType.LineMesh, triangleIndices);
            
            // Vector indices
            if (showVector)
            {
                triangleIndices = new uint[pCount * 2];
                for (uint i = 0; (i < pCount) && (i < uint.MaxValue); i++)
                {
                    triangleIndices[2 * i] = i;
                    triangleIndices[2 * i + 1] = (uint)(i + pCount);
                }
                obj.AddMeshIndices("vector", SlimDXRenderSystem.MeshType.LineMesh, triangleIndices);
            }

        }

        /// <summary>
        /// Initialize index and vertex buffers for PolyLine type mesh
        /// </summary>
        /// <param name="index">Index of mesh</param>
        /// <param name="bb">Bounding box of mesh</param>
        public void InitVBPolyLine()
        {
            int pCount = points.Length;

            Vector3[] vertices = new Vector3[pCount];
            Vector3[] vernormals = new Vector3[pCount];
            int[] vercolors = new int[pCount];


            for (int i = 0; i < pCount; i++)
            {
                Point3D point = (Point3D)points[i];

                Point3D normal;
                if (normals == null)
                    normal = new Point3D();
                else
                    normal = (Point3D)normals[i];

                vertices[i] = new Vector3((float)point.X, (float)point.Y, (float)point.Z);
                vernormals[i] = new Vector3((float)-normal.X, (float)-normal.Y, (float)-normal.Z);

                if (showColor)
                    vercolors[i] = ((ColorRGBA)(colors[i])).GetInt();
                else
                    vercolors[i] = new ColorRGBA(0.5f, 0.5f, 0.5f).GetInt();
            }
            obj = SceneManager.Instance.CreateObject(name);
            obj.SetMeshVertices(vertices, vernormals, vercolors, pCount);
        }

        /*public void InitIBPolyLine()
        {
            int lCount = points.Count;
            uint[] l = new uint[lCount * 2];
            uint k = 0;
            for (uint i = 0; i < primitives.Count; i++)
            {
                PolyLine line = (PolyLine)primitives[(int)i];
                for (uint j = 0; j < line.PointCount - 2; j++)
                {
                    l[2 * k] = (uint)line[(int)j];
                    l[2 * k + 1] = (uint)line[(int)j + 1];
                    k++;
                }
            }

            obj.AddMeshIndices("main", SlimDXRenderSystem.MeshType.LineMesh, l);
            obj.EnableDrawIndices("main");
        }*/


        /// <summary>
        /// Initialize index and vertex buffers for PolyLine type mesh
        /// </summary>
        /// <param name="index">Index of mesh</param>
        /// <param name="bb">Bounding box of mesh</param>
        public void InitVBLine()
        {
            int pCount;

            Vector3[] vertices;
            Vector3[] vernormals;
            int[] vercolors;

            if (showEdgeColor)
                pCount = primitives.Length * 2;
            else
                pCount = points.Length;

            vertices = new Vector3[pCount];
            vernormals = new Vector3[pCount];
            vercolors = new int[pCount];

            /*if (showEdgeColor)
            {
                for (int i = 0; i < primitives.Length; i++)
                {
                    Line line = (Line)primitives[(int)i];

                    int v1 = line.V1;
                    int v2 = line.V2;

                    Point3D point1 = (Point3D)points[v1];
                    Point3D point2 = (Point3D)points[v2];

                    Normal3D normal1;
                    Normal3D normal2;

                    if (normals == null)
                    {
                        normal1 = new Normal3D();
                        normal2 = new Normal3D();
                    }
                    else
                    {
                        normal1 = (Normal3D)normals[v1];
                        normal2 = (Normal3D)normals[v2];
                    }

                    vertices[2 * i] = new Vector3((float)point1.X, (float)point1.Y, (float)point1.Z);
                    vertices[2 * i + 1] = new Vector3((float)point2.X, (float)point2.Y, (float)point2.Z);
                    vernormals[2 * i] = new Vector3((float)-normal1.Nx, (float)-normal1.Ny, (float)-normal1.Nz);
                    vernormals[2 * i + 1] = new Vector3((float)-normal2.Nx, (float)-normal2.Ny, (float)-normal2.Nz);

                    int color = ((ColorRGBA)(edgeColors[i])).GetInt();
                    vercolors[2 * i] = color;
                    vercolors[2 * i + 1] = color;
                }
            }
            else*/
            {
                for (int i = 0; i < pCount; i++)
                {
                    Point3D point = (Point3D)points[i];

                    Point3D normal;
                    if (normals == null)
                        normal = new Point3D();
                    else
                        normal = (Point3D)normals[i];

                    vertices[i] = new Vector3((float)point.X, (float)point.Y, (float)point.Z);
                    vernormals[i] = new Vector3((float)-normal.X, (float)-normal.Y, (float)-normal.Z);

                    if (showColor)
                        vercolors[i] = ((ColorRGBA)(colors[i])).GetInt();
                    else
                        vercolors[i] = new ColorRGBA(0.5f, 0.5f, 0.5f).GetInt();
                }
            }
            obj = SceneManager.Instance.CreateObject(name);
            obj.SetMeshVertices(vertices, vernormals, vercolors, pCount);
        }

        /*public void InitIBLine()
        {
            int lCount = primitives.Length;
            uint[] l = new uint[lCount * 2];

            if (showEdgeColor)
            {
                for (uint i = 0; i < primitives.Length * 2; i++)
                {
                    l[i] = i;
                }
            }
            else
            {
                for (uint i = 0; i < primitives.Count; i++)
                {
                    Line line = (Line)primitives[(int)i];
                    l[2 * i] = (uint)line.V1;
                    l[2 * i + 1] = (uint)line.V2;
                }
            }

            obj.AddMeshIndices("main", SlimDXRenderSystem.MeshType.LineMesh, l);
            obj.EnableDrawIndices("main");
        }*/


        internal int GetPointNearestTo(Point3D c)
        {
            //Point3D c = new Point3D(cursorPointInvTransformed.X, cursorPointInvTransformed.Y, cursorPointInvTransformed.Z);
            double dist = 0;
            if (pointsTreeReady)
            {
                return (pointsTree.findNearest(c, out dist));
            }
            else
            {
                double minDist = double.MaxValue;
                int result = 0;

                for (int i = 0; i < this.points.Length; i++)
                {
                    Point3D p = points[i];

                    Point3D dif = p - c;
                    double distance = dif.dotProduct(dif);

                    if (distance < minDist)
                    {
                        result = i;
                        minDist = distance;
                    }
                }
                return (result);
            }
        }
    }
}
