using System;
using System.IO;
using System.Text;
using System.Collections;
using Framework;

namespace MeshBrowser.SlimDXRenderer
{
	/// <summary>
	/// Class for reading binary STL file format.
	/// </summary>
	public class ReaderBinStl
	{
		private int numberOfCells     = 0;
		private int numberOfVertices  = 0;

		private DataSet dataSet;

		private ArrayList cellNormalsArray = new ArrayList();
		private ArrayList cellArray        = new ArrayList();

		private Hashtable verticesHash     = new Hashtable();

        ProgressInfo progress;

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName, ProgressInfo progress)
		{
            this.progress = progress;
			FileStream fileStream = new FileStream(fileName, FileMode.Open, FileAccess.Read);

			BinaryReader binReader = new BinaryReader(fileStream);

			binReader.BaseStream.Position = 80;            // Skip input file header.
			this.numberOfCells = (int)binReader.ReadUInt32();

			try
			{
				for(int i = 0; i < this.numberOfCells; i++)
				{
					this.ReadTriangleNormal(binReader);

					this.ReadTriangleVertex(binReader);

					binReader.BaseStream.Position += 2;    // Skip 2 byte cell attributes.
				}
			}
			catch
			{
				long position = binReader.BaseStream.Position;
				fileStream.Close();
		
				throw new Exception("Error in input file on byte "+ position + ".");
			}

			fileStream.Close();                            // Close input file.

			this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
			                                               // Creates DataSet object.
			try
			{
				writeDataToDataSet();
			}
			catch
			{
				binReader.Close();
                throw new Exception("Unexpected error in input file. Check input file structure.");
			}

			return this.dataSet;
		}

		/// <summary>
		/// Reads one triangle normal from input file.
		/// </summary>
		/// <param name="binReader">Reference actual input file position.</param>
		private void ReadTriangleNormal(BinaryReader binReader)
		{
			Point3D normal;

			normal = new Point3D(binReader.ReadSingle(),
														binReader.ReadSingle(),
														binReader.ReadSingle());

			//this.cellNormalsArray.Add(normal);
		}

		/// <summary>
		/// Reads tree vertices of triangle from input file.
		/// </summary>
		/// <param name="binReader">Reference actual input file position.</param>
		private void ReadTriangleVertex(BinaryReader binReader)
		{
			Point3D  vertex1;
			Point3D  vertex2;
			Point3D  vertex3;
			Triangle triangle;

			vertex1 = new Point3D(binReader.ReadSingle(),
			                                            binReader.ReadSingle(),
			                                            binReader.ReadSingle());
            //vertex1.swapYZ();

			vertex2 = new Point3D(binReader.ReadSingle(),
				                                        binReader.ReadSingle(),
				                                        binReader.ReadSingle());
            //vertex2.swapYZ();

			vertex3 = new Point3D(binReader.ReadSingle(),
				                                        binReader.ReadSingle(),
				                                        binReader.ReadSingle());
            //vertex3.swapYZ();

			triangle = StoreVertexAndGetTriangle(vertex1, vertex3, vertex2);

			this.cellArray.Add(triangle);

            if (cellArray.Count%10000 == 0)
                progress((double)binReader.BaseStream.Position / binReader.BaseStream.Length);
		}

		/// <summary>
		/// Finds indexes of triangle's vertices.
		/// </summary>
		/// <param name="vertex1">First vertex of triangle.</param>
		/// <param name="vertex2">Second vertex of triangle.</param>
		/// <param name="vertex3">Third vertex of triangle.</param>
		/// <returns>Indexes definition of triangle.</returns>
		private Triangle StoreVertexAndGetTriangle(Point3D vertex1,
                                                             			 Point3D vertex2,
			                                                             Point3D vertex3)
		{
			Triangle triangle = new Triangle(0, 0, 0);

			if (this.numberOfVertices == 0)                 // Stores first three vertices.
			{
				this.verticesHash.Add(vertex1, 0);
				this.verticesHash.Add(vertex2, 1);
				this.verticesHash.Add(vertex3, 2);
				this.numberOfVertices = 3;

				triangle.V1 = 0;
				triangle.V2 = 1;
				triangle.V3 = 2;

				return triangle;
			}

			if (this.verticesHash.ContainsKey(vertex1))    // Checks if exist first vertex in vertices array.
			{                                              // True - assigns index to triangle.
				triangle.V1 = (int)this.verticesHash[vertex1];
			}
			else
			{                                              // False - adds new vertex to array.
				this.verticesHash.Add(vertex1, this.numberOfVertices);
				triangle.V1 = this.numberOfVertices;       // Assigns index (current number of vertices) to triangle.
				this.numberOfVertices++;
			}

			if (this.verticesHash.ContainsKey(vertex2))    // Same as first vertex.
			{
				triangle.V2 = (int)this.verticesHash[vertex2];
			}
			else
			{
				this.verticesHash.Add(vertex2, this.numberOfVertices);
				triangle.V2 = this.numberOfVertices;
				this.numberOfVertices++;
			}

			if (this.verticesHash.ContainsKey(vertex3))    // Same as first vertex.
			{
				triangle.V3 = (int)this.verticesHash[vertex3];
			}
			else
			{
				this.verticesHash.Add(vertex3, this.numberOfVertices);
				triangle.V3 = this.numberOfVertices;
				this.numberOfVertices++;
			}

			return triangle;
		}

		/// <summary>
		/// Writes readed data to DataSet.
		/// </summary>
		private void writeDataToDataSet()
		{
			int   index = 0;

			foreach (Point3D point3d in this.verticesHash.Keys)
			{                                              // Adds vertices.
				this.dataSet.AddVertex((int)this.verticesHash[point3d] , point3d);
			}

			/*index = 0;
			foreach (Zcu.Mve.Visualization.Normal3D normal3d in this.cellNormalsArray)
			{                                              // Adds cell normals.
				this.dataSet.AddCellNormal(index, normal3d);
				index++;
			}*/

			index = 0;
			foreach (Triangle triangle in this.cellArray)
			{                                              // Adds cells (triangles).
				this.dataSet.AddTriangle(index, triangle);
				index++;
			}
		}
	}
}
