using System;
using System.IO;
using System.Text;
using System.Collections.Generic;


using Framework;

namespace MeshBrowser.SlimDXRenderer
{
    /// <summary>
    /// Class for reading PLY file format.
    /// </summary>
    public class ReaderPly
    {
        private int numberOfCells = 0;
        private int numberOfVertices = 0;

        private int currentLine = 0;                        // Actual line of readed file.

        private DataSet dataSet;

        private string exception = null;                   // Exception error string.

        private enum Format { ascii, binary_big_endian, binary_little_endian };
        private enum ReadingStage { none, vertex, face };

        System.Globalization.NumberFormatInfo nfi;

        private int TypeSize(string typeName)
        {
            switch (typeName)
            {
                case "char":
                case "uchar":
                case "int8":
                case "uint8":
                    return 1;

                case "short":
                case "ushort":
                case "int16":
                case "uint16":
                    return 2;

                case "int":
                case "uint":
                case "int32":
                case "uint32":
                case "float":
                    return 4;

                case "double":
                    return 8;
            }
            return 0;
        }

        ProgressInfo progress;

        /// <summary>
        /// Reads input file.
        /// </summary>
        /// <param name="fileName">Full filename.</param>
        /// <returns>Filled object of DataSet class.</returns>
        public DataSet ReadFile(string fileName, ProgressInfo progress)
        {
            this.progress = progress;
            nfi = new System.Globalization.NumberFormatInfo();
            nfi.NumberDecimalSeparator = ".";            // Setting of decimal separator.

            StreamReader streamReader = new StreamReader(fileName);

            string line = null;
            byte numberOf = 2;                           //  Counting parts of file head.

            Format format = Format.ascii;
            ReadingStage stage = ReadingStage.none;

            int vertexDataLength = 0;
            int vertexCoordinateLength = 0;
            int faceIndexCountLength = 0;
            int faceIndexLength = 0;
            int discardBefore = 0;
            int discardAfter = 0;
            int headerLength = 0;

            try
            {
                List<PropertyInfo> PropertyInfoList = new List<PropertyInfo>();
                while (((line = streamReader.ReadLine()) != null) && (!line.StartsWith("end_header")))
                {                                          // File head processing.
                    this.currentLine++;
                    headerLength += line.Length + 1;


                    if (line.StartsWith("element ver"))
                    {                                      // Reads number of vertices.
                        this.numberOfVertices = this.IntNumber(line);
                        stage = ReadingStage.vertex;
                        numberOf--;
                    }
                    else if (line.StartsWith("element fac"))
                    {                                      // Reads number of cells.
                        this.numberOfCells = this.IntNumber(line);
                        stage = ReadingStage.face;
                        numberOf--;
                    }
                    else if (line.StartsWith("format"))
                    {
                        string[] parts = line.Split(new char[] { ' ' });
                        if (parts[1] == "binary_big_endian")
                        {
                            format = Format.binary_big_endian;
                        }
                        else if (parts[1] == "binary_little_endian")
                        {
                            format = Format.binary_little_endian;
                        }
                        else
                        {
                            format = Format.ascii;
                        }
                    }
                    else if (line.StartsWith("element"))
                    {
                        stage = ReadingStage.none;
                    }
                    else if (line.StartsWith("property"))
                    {
                        switch (stage)
                        {
                            case ReadingStage.vertex:
                                string[] parts = line.Split(new char[] { ' ' });
                                if (vertexDataLength == 0)
                                {
                                    switch (parts[1])
                                    {
                                        case "float":
                                        case "float32":
                                            vertexCoordinateLength = 4;
                                            break;

                                        case "double":
                                        case "float64":
                                            vertexCoordinateLength = 8;
                                            break;
                                    }
                                }
                                switch (parts[1])
                                {
                                    case "float":
                                    case "float32":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(float)));
                                        break;

                                    case "double":
                                    case "float64":
                                        vertexDataLength += 8;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(double)));
                                        break;

                                    case "uint":
                                    case "uint32":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(uint)));
                                        break;

                                    case "int":
                                    case "int32":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(int)));
                                        break;

                                    case "ushort":
                                    case "uint16":
                                        vertexDataLength += 2;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(ushort)));
                                        break;

                                    case "short":
                                    case "int16":
                                        vertexDataLength += 2;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(short)));
                                        break;

                                    case "uchar":
                                    case "uint8":
                                        vertexDataLength += 1;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(byte)));
                                        break;

                                    case "char":
                                    case "int8":
                                        vertexDataLength += 1;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(sbyte)));
                                        break;
                                }
                                break;
                            case ReadingStage.face:
                                parts = line.Split(new char[] { ' ' });
                                if (parts[1].Equals("list"))
                                {
                                    faceIndexCountLength += TypeSize(parts[2]);
                                    faceIndexLength += TypeSize(parts[3]);
                                }
                                else
                                {
                                    if (faceIndexLength == 0) discardBefore += TypeSize(parts[1]);
                                    else discardAfter += TypeSize(parts[2]);
                                }
                                break;
                        }
                    }

                    if (numberOf == 0)
                    {                                      // Head reading is done, creating DataSet object.
                        this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
                        this.dataSet.vertexPropertyInfoList = PropertyInfoList;
                    }
                }

                headerLength += line.Length + 1;

                if (numberOf != 0)                             // If number of triangles or vertices wasn't found - throws exception.
                {
                    this.exception = "Invalid file format - can't find number of triangles and/or number of vertices.";
                    throw new Exception("Invalid file format - can't find number of triangles and/or number of vertices.");
                }

                /*dataSet.vertexAttrs = new UniformDataArray[dataSet.vertexPropertyInfoList.Count - 3];
                for (int i = 0; i < (dataSet.vertexPropertyInfoList.Count - 3); i++)
                {
                    Type type = dataSet.vertexPropertyInfoList[i + 3].type;
                    if ((type == typeof(byte)) || (type == typeof(sbyte)) || (type == typeof(int)) ||
                        (type == typeof(uint)) || (type == typeof(short)) || (type == typeof(ushort)))
                        dataSet.vertexAttrs[i] = new UniformDataArray(typeof(Integer), numberOfVertices);
                    else if ((type == typeof(float)) || (type == typeof(double)))
                        dataSet.vertexAttrs[i] = new UniformDataArray(typeof(Scalar), numberOfVertices);
                }*/

                if (format == Format.ascii)
                {
                    this.ReadVertices(ref streamReader);
                    this.ReadFaces(ref streamReader);
                }
                else
                {
                    this.ReadVerticesBinary(ref streamReader, format, vertexDataLength, vertexCoordinateLength, numberOfVertices, headerLength, dataSet.vertexPropertyInfoList);
                    this.ReadFacesBinary(ref streamReader, format, numberOfCells, faceIndexCountLength, faceIndexLength, discardBefore, discardAfter);
                }


            }
            catch (System.Exception e)
            {
                streamReader.Close();
                if (this.exception == null)
                {
                    throw new Exception("Error in input file on line " + this.currentLine + ".", e);
                }
                else
                {
                    throw new Exception(this.exception, e);
                }
            }

            streamReader.Close();

            return this.dataSet;
        }

        /// <summary>
        /// Reads integer number in input string.
        /// </summary>
        /// <param name="line">Input string.</param>
        /// <returns>Readed number.</returns>
        private int IntNumber(string line)
        {
            string[] elements;

            elements = line.Split(" :\t".ToCharArray());

            return Int32.Parse(elements[elements.Length - 1]);
        }

        /// <summary>
        /// Reads one vertex in input string.
        /// </summary>
        /// <param name="line">Input string.</param>
        /// <returns>Readed vertex (3D point).</returns>
        private Point3D ReadPoint3D(string line)
        {
            string[] elements;
            int elementsLength = 0;

            elements = line.Split(" :\t".ToCharArray());

            for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
            {
                if (elements[i].Trim() != "")
                {
                    elements[elementsLength++] = elements[i];
                }
            }

            return new Point3D(Double.Parse(elements[0], nfi),
                                                     Double.Parse(elements[1], nfi),
                                                     Double.Parse(elements[2], nfi));
        }

        /// <summary>
        /// Reads all vertices from input file.
        /// </summary>
        /// <param name="streamReader">Reference to input file.</param>
        private void ReadVertices(ref StreamReader streamReader)
        {

            Point3D vertex;
            string line;
            int i = 0;                                  // Index of actual vertex.

            while ((line = streamReader.ReadLine()) != null)
            {
                this.currentLine++;

                line = line.Trim();

                if (line == "")                            // Skips free line.
                {
                    continue;
                }
                else                                       // Trys to read one vertex.
                {
                    vertex = this.ReadPoint3D(line);
                    this.dataSet.AddVertex(i, vertex);

                    if (++i == numberOfVertices)           // Number of vertices reached - exit.
                    {
                        break;
                    }
                }
                if (currentLine % 10000 == 0)
                    progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
            }
        }

        /// <summary>
        /// Reads one polygon in input string.
        /// </summary>
        /// <param name="line">Input string.</param>
        /// <returns>Readed polygon.</returns>
        private Triangle ReadFace(string line)
        {
            string[] elements;
            int[] indices;
            int elemCount = 0;
            int idCount;

            elements = line.Split(" :\t".ToCharArray());

            for (int i = 0; i < elements.Length; i++)      // Deletes elements that contains spaces only.
                if (elements[i].Trim() != "")
                    elements[elemCount++] = elements[i];


            idCount = Int32.Parse(elements[0]);      // Reads number of polygon vertices.
            indices = new int[idCount];              // Allocate memory for polygon.

            for (int i = 0; i < idCount; i++)        // Reads indexes of polygon vertices.
            {
                indices[i] = Int32.Parse(elements[i + 1]);

                if (indices[i] >= this.numberOfVertices)
                {
                    this.exception = "Error in input file on line " + this.currentLine + " - index is greater then number of vertices.";
                    throw new Exception("Error in input file on line " + this.currentLine + " - index is greater then number of vertices.");
                }
            }

            return new Triangle(indices[0], indices[1], indices[2]);

            /*if (idCount == 3) return new Triangle(indices[0], indices[1], indices[2]);
            else return new Polygon(indices);*/
        }

        /// <summary>
        /// Reads all polygons from input file.
        /// </summary>
        /// <param name="streamReader">Reference to input file.</param>
        private void ReadFaces(ref StreamReader streamReader)
        {
            Triangle face;
            string line;
            int i = 0;                                  // Index of actual triangle.

            while ((line = streamReader.ReadLine()) != null)
            {
                this.currentLine++;

                line = line.Trim();

                if (line == "")                            // Skips free line.
                {
                    continue;
                }
                else                                       // Try to reads one triangle.
                {
                    face = this.ReadFace(line);

                    //if (face is Triangle) 
                        this.dataSet.AddTriangle(i, (Triangle)face);
                    //else this.dataSet.AddPolygon(i, (Polygon)face);

                    if (++i == numberOfCells)              // Number of triangles reached - exit.
                    {
                        break;
                    }
                }
                if (currentLine % 10000 == 0)
                    progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
            }

            if (i != this.numberOfCells)
            {
                this.exception = "Error in input file - number of cells and/or vertices doesn't match.";
                throw new Exception("Error in input file - number of cells and/or vertices doesn't match.");
            }
        }

        private byte[] reverse4(byte[] original, int pos)
        {
            byte[] result = new byte[4];
            for (int i = 0; i < 4; i++)
                result[i] = original[pos + 3 - i];
            return result;
        }

        private byte[] reverse2(byte[] original, int pos)
        {
            byte[] result = new byte[2];
            for (int i = 0; i < 2; i++)
                result[i] = original[pos + 1 - i];
            return result;
        }

        private byte[] reverse8(byte[] original, int pos)
        {
            byte[] result = new byte[8];
            for (int i = 0; i < 8; i++)
                result[i] = original[pos + 7 - i];
            return result;
        }

        private void ReadVerticesBinary(ref StreamReader streamReader, Format format, int vertexDataLength, int vertexCoordinateLength, int vertexCount, int headerLength, List<PropertyInfo> piList)
        {
            BinaryReader br = new BinaryReader(streamReader.BaseStream);
            br.BaseStream.Seek(headerLength, SeekOrigin.Begin);
            if (format == Format.binary_big_endian)
            {
                for (int i = 0; i < vertexCount; i++)
                {
                    byte[] buffer = br.ReadBytes(vertexDataLength);
                    double[] vertex = new double[3];
                    if (vertexCoordinateLength == 4)
                    {

                        vertex[0] = BitConverter.ToSingle(reverse4(buffer, 0), 0);
                        vertex[1] = BitConverter.ToSingle(reverse4(buffer, 4), 0);
                        vertex[2] = BitConverter.ToSingle(reverse4(buffer, 8), 0);
                        //int pos = 4 * 3;
                        //ReadProps(i, buffer, pos);
                    }
                    else if (vertexCoordinateLength == 8)
                    {
                        vertex[0] = BitConverter.ToDouble(reverse8(buffer, 0), 0);
                        vertex[1] = BitConverter.ToDouble(reverse8(buffer, 8), 0);
                        vertex[2] = BitConverter.ToDouble(reverse8(buffer, 16), 0);
                        //int pos = 8 * 3;
                        //ReadProps(i, buffer, pos);
                    }
                    this.dataSet.AddVertex(i, new Point3D(vertex[0], vertex[1], vertex[2]));
                    if (i % 10000 == 0)
                        progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
                }
            }
            else // little endian
            {
                for (int i = 0; i < vertexCount; i++)
                {
                    byte[] buffer = br.ReadBytes(vertexDataLength);
                    double[] vertex = new double[3];
                    if (vertexCoordinateLength == 4)
                    {

                        vertex[0] = BitConverter.ToSingle(buffer, 0);
                        vertex[1] = BitConverter.ToSingle(buffer, 4);
                        vertex[2] = BitConverter.ToSingle(buffer, 8);
                        int pos = 4 * 3;
                        ReadProps(i, buffer, pos);
                    }
                    else if (vertexCoordinateLength == 8)
                    {
                        vertex[0] = BitConverter.ToDouble(buffer, 0);
                        vertex[1] = BitConverter.ToDouble(buffer, 8);
                        vertex[2] = BitConverter.ToDouble(buffer, 16);
                        int pos = 8 * 3;
                        ReadProps(i, buffer, pos);
                    }
                    this.dataSet.AddVertex(i, new Point3D(vertex[0], vertex[1], vertex[2]));
                    if (i % 10000 == 0)
                        progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
                }
            }
        }

        private int ReadProps(int i, byte[] buffer, int pos)
        {
            if (dataSet.vertexAttrs != null)
            {
                for (int j = 0; j < dataSet.vertexAttrs.Length; j++)
                {
                    byte[] attr = (byte[])dataSet.vertexAttrs[j];
                    //if (attr[0] is byte)
                    //{
                    attr[i] = buffer[pos];
                    pos++;
                    //}
                }
                return pos;
            }
            else return pos;
        }
        private void ReadFacesBinary(ref StreamReader streamReader, Format format, int faceCount, int listLengthLength, int ListItemLength, int discardBytesBefore, int discardBytesAfter)
        {
            BinaryReader br = new BinaryReader(streamReader.BaseStream);
            if (format == Format.binary_big_endian)
            {
                for (int i = 0; i < faceCount; i++)
                {
                    if (discardBytesBefore > 0) br.ReadBytes(discardBytesBefore);

                    byte[] listLength = br.ReadBytes(listLengthLength);
                    int length = 0;
                    if (listLengthLength == 1)
                        length = listLength[0];
                    if (listLengthLength == 2)
                        length = BitConverter.ToUInt16(listLength, 0);
                    int[] indices = new int[length];

                    byte[] buffer = br.ReadBytes(length * ListItemLength);

                    if (discardBytesAfter > 0) br.ReadBytes(discardBytesAfter);

                    for (int j = 0; j < length; j++)
                    {
                        if (ListItemLength == 2)
                            indices[j] = BitConverter.ToUInt16(reverse2(buffer, j * ListItemLength), 0);
                        else if (ListItemLength == 4)
                            indices[j] = (int)BitConverter.ToUInt32(reverse4(buffer, j * ListItemLength), 0);
                    }
                    //if (length == 3)
                        this.dataSet.AddTriangle(i, new Triangle(indices[0], indices[1], indices[2]));
                    //else
                    //    this.dataSet.AddPolygon(i, new Polygon(indices));
                    if (i % 10000 == 0)
                            progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
                }
            }
            else
            {
                for (int i = 0; i < faceCount; i++)
                {
                    if (discardBytesBefore > 0) br.ReadBytes(discardBytesBefore);

                    byte[] listLength = br.ReadBytes(listLengthLength);
                    int length = 0;
                    if (listLengthLength == 1)
                        length = listLength[0];
                    if (listLengthLength == 2)
                        length = BitConverter.ToUInt16(listLength, 0);
                    int[] indices = new int[length];

                    byte[] buffer = br.ReadBytes(length * ListItemLength);

                    if (discardBytesAfter > 0) br.ReadBytes(discardBytesAfter);

                    for (int j = 0; j < length; j++)
                    {
                        if (ListItemLength == 2)
                            indices[j] = BitConverter.ToUInt16(buffer, j * ListItemLength);
                        else if (ListItemLength == 4)
                            indices[j] = (int)BitConverter.ToUInt32(buffer, j * ListItemLength);
                    }
                    //if (length == 3)
                        this.dataSet.AddTriangle(i, new Triangle(indices[0], indices[1], indices[2]));
                    //else
                    //    this.dataSet.AddPolygon(i, new Polygon(indices));
                    if (i % 10000 == 0)
                            progress((double)streamReader.BaseStream.Position / streamReader.BaseStream.Length);
                }
            }
        }
    }
}
