﻿using System;
using System.ComponentModel;
using System.Threading;
using System.Drawing;
using SlimDXRenderSystem;
using Framework;
using System.Windows;

namespace MeshBrowser.SlimDXRenderer
{
    /// <summary>
    /// SlimDX DirectX based renderer for Triangle meshes
    /// </summary>
    public class SlimDXRenderer
    {
        #region Fields
        private ManualResetEvent windowDisplayed;
        private Window1 renderWindow;
        private D3DVersion d3dVersion = D3DVersion.Direct3D9;
        
        private int meshCount = 1;

        private System.Drawing.Color backgroundColor = System.Drawing.Color.FromArgb(0,0,0);

        private bool sameConnectivity = false;
        private bool showColorAttribute = false;
        private bool showVectorAttribute = false;
        private bool showEdgeColorAttribute = false;

        #region Properties
        /// <summary>
        /// Version of Direct3D to be used.
        /// </summary>
        [CategoryAttribute("Device"), Description("Version of Direct3D to be used.")]
        public D3DVersion D3dVersion
        {
            get { return (d3dVersion); }
            set { d3dVersion = value; }
        }                

        /// <summary>
        /// Determines color of window background
        /// </summary>
        [CategoryAttribute("Window"), Description("Determines color of window background.")]
        public System.Drawing.Color BackgroundColor
        {
            get { return (backgroundColor); }
            set { backgroundColor = value; }
        }

        /// <summary>
        /// Determines if the input meshes have the same connectivity or not. If set to true, the only the connectivity information form the first mesh will be used. All remaining meshes must have the same number of vertices. 
        /// Memory usage is lower when set to true.
        /// </summary>
        [CategoryAttribute("Mesh"), Description("Determines weher the input meshes have the same connectivity or not. If set to true, the only the connectivity information form the first mesh will be used. All remaining meshes must have the same number of vertices. Memory usage is lower when set to true.")]
        public bool SameConnectivity
        {
            get { return (sameConnectivity); }
            set { sameConnectivity = value; }
        }

        /// <summary>
        /// Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.
        /// </summary>
        [CategoryAttribute("Attribute Setting"), Description("Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.")]
        public bool ShowColorAttribute
        {
            get
            {
                return (showColorAttribute);
            }
            set
            {
                showColorAttribute = value;
            }
        }

        /// <summary>
        /// Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.
        /// </summary>
        [CategoryAttribute("Attribute Setting"), Description("Determines wether an attribute of ColorRGBA type will be used as color source of the edges for the object.")]
        public bool ShowEdgeColorAttribute
        {
            get { return showEdgeColorAttribute; }
            set { showEdgeColorAttribute = value; }
        }


        /// <summary>
        /// Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.
        /// </summary>
        [CategoryAttribute("Attribute Setting"), Description("Determines wether an attribute of PointND type will be displayed. Set the vectorAttributeName to specify which attribute to be used.")]
        public bool ShowVectorAttribute
        {
            get
            {
                return (showVectorAttribute);
            }
            set
            {
                showVectorAttribute = value;
            }
        }
        #endregion
        #endregion

        /// <summary>
        /// Constructor of SlimDXRenderer class
        /// </summary>
        public SlimDXRenderer()
		{
            windowDisplayed = new ManualResetEvent(false);
		}        

        /// <summary>
        /// Method loads the meshes and display the rendering window
        /// </summary>
        public void ShowRenderWindow(TriangleMesh[] meshes)
        {
            meshCount = meshes.Length;
            renderWindow = new Window1(meshes, showColorAttribute, showVectorAttribute);
            Application app = new Application();
            app.Run(renderWindow);
            renderWindow = null;
            windowDisplayed.Set();
        }
    }
}
